"""
Velmu Commands Extension

Provides command handling for Velmu bots.

Example
-------
```python
from velmu.ext import commands

bot = commands.Bot(command_prefix='+')

@bot.command()
async def ping(ctx):
    await ctx.send('Pong!')

bot.run('TOKEN')
```
"""

from .core import (
    Bot,
    Context,
    Command,
    Cog,
    command,
    cooldown,
    CooldownMapping
)

from ...errors import (
    CommandError,
    MissingRequiredArgument,
    BadArgument,
    CheckFailure,
    CommandNotFound,
    CommandOnCooldown
)

from .checks import check, has_permissions, is_owner
from .converters import (
    Converter, 
    MemberConverter, 
    ChannelConverter,
    UserConverter,
    RoleConverter,
    GuildConverter,
    MessageConverter,
    InviteConverter
)

__all__ = [
    'Bot',
    'Context',
    'Command',
    'Cog',
    'command',
    'cooldown',
    'CooldownMapping',
    'check',
    'has_permissions',
    'is_owner',
    'Converter',
    'MemberConverter',
    'ChannelConverter',
    'UserConverter',
    'RoleConverter',
    'GuildConverter',
    'MessageConverter',
    'InviteConverter',
    'CommandError',
    'MissingRequiredArgument',
    'BadArgument',
    'CheckFailure',
    'CommandOnCooldown',
]



"""
velmu.models.base
~~~~~~~~~~~~~~~~~

Base model class.
"""

class BaseModel:
    """Base class for all Velmu models."""
    
    __slots__ = ('_state', 'id')

    def __init__(self, state, data):
        self._state = state
        self.id = data.get('id')

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)



"""
Velmu Commands Framework

Extension for creating bot commands with decorators.
"""
import inspect
import asyncio
import time
from functools import wraps
from typing import Optional, Dict, Any, Callable, List

from ...models import Message, User
from ...client import Client
from ...errors import (
    CommandError,
    MissingRequiredArgument,
    BadArgument,
    CheckFailure,
    CommandOnCooldown
)


class CooldownMapping:
    """Tracks cooldowns per user/channel/guild."""
    
    def __init__(self, rate: int, per: float, type_: str = 'user'):
        self.rate = rate  # Number of uses
        self.per = per  # Per X seconds
        self.type = type_  # 'user', 'channel', 'guild'
        self._cache: Dict[str, List[float]] = {}

    def _get_key(self, ctx) -> str:
        if self.type == 'user':
            return ctx.author.id
        elif self.type == 'channel':
            return ctx.channel.id
        elif self.type == 'guild':
            return ctx.guild.id if ctx.guild else ctx.channel.id
        return ctx.author.id

    def get_retry_after(self, ctx) -> Optional[float]:
        """Check if on cooldown, return seconds to wait or None."""
        key = self._get_key(ctx)
        now = time.time()
        
        if key not in self._cache:
            self._cache[key] = []
        
        # Clean old entries
        self._cache[key] = [t for t in self._cache[key] if now - t < self.per]
        
        if len(self._cache[key]) >= self.rate:
            oldest = min(self._cache[key])
            return self.per - (now - oldest)
        
        return None

    def update(self, ctx):
        """Record a command use."""
        key = self._get_key(ctx)
        if key not in self._cache:
            self._cache[key] = []
        self._cache[key].append(time.time())


class Command:
    """Represents a bot command.
    
    Attributes
    ----------
    name : str
        The command name.
    callback : Callable
        The command function.
    help : str
        Help text for the command.
    aliases : list[str]
        Alternative names for the command.
    checks : list[Callable]
        Check functions that must pass.
    cooldown : CooldownMapping
        Cooldown tracking for the command.
    """
    
    def __init__(self, func, **kwargs):
        self.callback = func
        self.name = kwargs.get('name') or func.__name__
        
        # Docstring parsing
        doc = inspect.getdoc(func)
        self.help = kwargs.get('help') or doc or ''
        self.description = kwargs.get('description') or (doc.splitlines()[0] if doc else '')
        
        self.category = kwargs.get('category')
        self.aliases = kwargs.get('aliases', [])
        self.checks = kwargs.get('checks', [])
        self.cooldown: Optional[CooldownMapping] = None
        self.cog = None  # Set when added to a Cog

    @property
    def signature(self) -> str:
        """Returns a string representing the command signature (usage)."""
        target_func = getattr(self.callback, '__original_func__', self.callback)
        params = list(inspect.signature(target_func).parameters.values())
        
        # Skip self/ctx
        skip_count = 0
        if params and params[0].name == 'self': skip_count += 1
        if len(params) > skip_count and params[skip_count].name in ('ctx', 'context'): skip_count += 1
        params = params[skip_count:]
        
        sig_parts = []
        for param in params:
            name = param.name
            if param.kind == param.VAR_POSITIONAL:
                sig_parts.append(f"[{name}...]")
            elif param.kind == param.KEYWORD_ONLY:
                sig_parts.append(f"<{name}...>")
            elif param.default is not param.empty:
                sig_parts.append(f"[{name}={param.default}]" if param.default is not None else f"[{name}]")
            else:
                sig_parts.append(f"<{name}>")
                
        return " ".join(sig_parts)

    async def invoke(self, ctx):
        """Invoke the command with the given context."""
        # Run checks
        for check in self.checks:
            try:
                result = check(ctx) if not asyncio.iscoroutinefunction(check) else await check(ctx)
                if not result:
                    raise CheckFailure(f'Check failed for command {self.name}')
            except Exception as e:
                raise CheckFailure(str(e))
        
        # Check cooldown
        if self.cooldown:
            retry_after = self.cooldown.get_retry_after(ctx)
            if retry_after:
                raise CommandOnCooldown(retry_after)
            self.cooldown.update(ctx)
        
        # Parse arguments
        target_func = getattr(self.callback, '__original_func__', self.callback)
        sig = inspect.signature(target_func)
        params = list(sig.parameters.values())

        # Skip 'self' and 'ctx'
        skip_count = 0
        if params and params[0].name == 'self':
            skip_count += 1
        if len(params) > skip_count and params[skip_count].name in ('ctx', 'context'):
            skip_count += 1
             
        params = params[skip_count:]
        
        converted_args = []
        remaining_args = ctx.args
        
        for i, param in enumerate(params):
            # Keyword-only argument (consumes rest as one string)
            if param.kind == param.KEYWORD_ONLY:
                val = ' '.join(remaining_args[i:])
                ctx.kwargs[param.name] = val
                break
            
            # Variable Positional (*args) (consumes rest as list)
            if param.kind == param.VAR_POSITIONAL:
                rest = remaining_args[i:]
                # Apply conversion if needed
                annotation = param.annotation
                if annotation is not param.empty:
                    converted_rest = []
                    from .converters import (
                        MemberConverter, UserConverter, RoleConverter, 
                        ChannelConverter, GuildConverter, MessageConverter, InviteConverter
                    )
                    
                    for item in rest:
                        if hasattr(annotation, 'convert'):
                            converter = annotation()
                            item = await converter.convert(ctx, item)
                        else:
                            type_name = getattr(annotation, '__name__', str(annotation))
                            converter = None
                            
                            if type_name == 'Member': converter = MemberConverter()
                            elif type_name == 'User': converter = UserConverter()
                            elif type_name == 'Role': converter = RoleConverter()
                            elif type_name == 'Channel': converter = ChannelConverter()
                            elif type_name == 'Guild': converter = GuildConverter()
                            elif type_name == 'Message': converter = MessageConverter()
                            elif type_name == 'Invite': converter = InviteConverter()
                            
                            if converter:
                                item = await converter.convert(ctx, item)
                            elif callable(annotation):
                                item = annotation(item)
                                
                        converted_rest.append(item)
                    converted_args.extend(converted_rest)
                else:
                    converted_args.extend(rest)
                break
            
            # Missing argument
            if i >= len(remaining_args):
                if param.default is not param.empty:
                    converted_args.append(param.default)
                    continue
                else:
                    raise MissingRequiredArgument(param)

            arg = remaining_args[i]
            

            # Type conversion
            annotation = param.annotation
            if annotation is not param.empty:
                try:
                    if hasattr(annotation, 'convert'):
                        # If the annotation itself is a Converter class
                        converter = annotation()
                        arg = await converter.convert(ctx, arg)
                    else:
                        # Check against known models mapping
                        from .converters import (
                            MemberConverter, UserConverter, RoleConverter, 
                            ChannelConverter, GuildConverter, MessageConverter, InviteConverter
                        )
                        
                        type_name = getattr(annotation, '__name__', str(annotation))
                        
                        converter = None
                        if type_name == 'Member':
                            converter = MemberConverter()
                        elif type_name == 'User':
                            converter = UserConverter()
                        elif type_name == 'Role':
                            converter = RoleConverter()
                        elif type_name == 'Channel':
                            converter = ChannelConverter()
                        elif type_name == 'Guild':
                            converter = GuildConverter()
                        elif type_name == 'Message':
                            converter = MessageConverter()
                        elif type_name == 'Invite':
                            converter = InviteConverter()
                        
                        if converter:
                            arg = await converter.convert(ctx, arg)
                        elif callable(annotation):
                            arg = annotation(arg)

                except Exception as e:
                    raise BadArgument(f'Could not convert "{arg}" to {getattr(annotation, "__name__", str(annotation))}')

            converted_args.append(arg)

        await self.callback(ctx, *converted_args, **ctx.kwargs)


def command(**kwargs):
    """Decorator to create a command.
    
    Parameters
    ----------
    name : str, optional
        The command name. Defaults to function name.
    help : str, optional
        Help text for the command.
    aliases : list[str], optional
        Alternative names for the command.
    
    Example
    -------
    ```python
    @commands.command(name='ping', help='Check bot latency')
    async def ping(ctx):
        await ctx.send('Pong!')
    ```
    """
    def decorator(func):
        return Command(func, **kwargs)
    return decorator


def cooldown(rate: int, per: float, type_: str = 'user'):
    """Decorator to add a cooldown to a command.
    
    Parameters
    ----------
    rate : int
        Number of uses allowed.
    per : float
        Time period in seconds.
    type_ : str
        'user', 'channel', or 'guild'
    
    Example
    -------
    ```python
    @commands.command()
    @commands.cooldown(1, 60, 'user')  # Once per minute per user
    async def daily(ctx):
        await ctx.send('You got your daily reward!')
    ```
    """
    def decorator(func):
        if isinstance(func, Command):
            func.cooldown = CooldownMapping(rate, per, type_)
        else:
            # Store for later when wrapped in Command
            if not hasattr(func, '__command_cooldown__'):
                func.__command_cooldown__ = CooldownMapping(rate, per, type_)
        return func
    return decorator


class Cog:
    """Base class for command groups (Cogs).
    
    Cogs allow you to organize commands into separate files/classes.
    
    Example
    -------
    ```python
    class General(commands.Cog):
        def __init__(self, bot):
            self.bot = bot
        
        @commands.command()
        async def ping(self, ctx):
            await ctx.send('Pong!')
    
    def setup(bot):
        bot.add_cog(General(bot))
    ```
    """
    
    def __init__(self, bot):
        self.bot = bot
        
    def get_commands(self) -> List[Command]:
        """Return all commands in this Cog."""
        cmds = []
        for name, member in inspect.getmembers(self):
            if isinstance(member, Command):
                cmds.append(member)
        return cmds

    @classmethod
    def listener(cls, name: str = None):
        """Decorator to register an event listener in a Cog.
        
        Parameters
        ----------
        name : str, optional
            Event name. Defaults to function name.
        
        Example
        -------
        ```python
        @commands.Cog.listener()
        async def on_message(self, message):
            print(f'Message: {message.content}')
        ```
        """
        def decorator(func):
            func.__is_listener__ = True
            func.__listener_name__ = name or func.__name__
            return func
        return decorator


class Context:
    """Context for a command invocation.
    
    Attributes
    ----------
    bot : Bot
        The bot instance.
    message : Message
        The message that triggered the command.
    author : User
        The message author.
    channel : Channel
        The channel the command was invoked in.
    guild : Guild
        The guild (if any).
    command : Command
        The command being invoked.
    prefix : str
        The prefix used.
    args : list[str]
        Parsed arguments.
    """
    
    def __init__(self, **kwargs):
        self.bot = kwargs.get('bot')
        self.message = kwargs.get('message')
        self.view = kwargs.get('view')
        self.prefix = kwargs.get('prefix', '')
        self.command: Optional[Command] = kwargs.get('command')
        self.invoked_with = kwargs.get('invoked_with', '')
        
        self.args: List[str] = []
        self.kwargs: Dict[str, Any] = {}
        
        self.guild = getattr(self.message, 'guild', None)
        self.channel = self.message.channel
        self.author = self.message.author
    
    @property
    def me(self) -> Optional[User]:
        """The bot's user object."""
        return self.bot.user

    async def send(self, content=None, embed=None, **kwargs) -> Message:
        """Send a message to the command channel.
        
        Parameters
        ----------
        content : str, optional
            Message content.
        embed : Embed or dict, optional
            Embed to send.
        
        Returns
        -------
        Message
            The sent message.
        """
        if embed and hasattr(embed, 'to_dict'):
            embed = embed.to_dict()
        return await self.channel.send(content, embed=embed, **kwargs)

    async def reply(self, content=None, embed=None, **kwargs) -> Message:
        """Reply to the command message.
        
        Parameters
        ----------
        content : str, optional
            Message content.
        embed : Embed or dict, optional
            Embed to send.
        
        Returns
        -------
        Message
            The sent message.
        """
        if embed and hasattr(embed, 'to_dict'):
            embed = embed.to_dict()
        return await self.message.reply(content, embed)


class Bot(Client):
    """Extended Client with command handling.
    
    Parameters
    ----------
    command_prefix : str or Callable
        The command prefix.
    **options
        Options passed to Client.
    
    Attributes
    ----------
    commands : dict
        Registered commands.
    cogs : dict
        Loaded Cogs.
    """
    
    def __init__(self, command_prefix, **options):
        super().__init__(**options)
        self.command_prefix = command_prefix
        self.commands: Dict[str, Command] = {}
        self.cogs: Dict[str, Cog] = {}
        self.extensions: Dict[str, Any] = {}
        
        # Register on_message listener
        self.event(self.on_message)

    async def close(self):
        """Closes the bot and unloads extensions."""
        await super().close()
        # Add any extension cleanup here if necessary

    # =========================================
    # Extension Loading
    # =========================================

    def load_extension(self, name: str):
        """Load an extension module.
        
        Parameters
        ----------
        name : str
            Module name (e.g., 'cogs.general').
        """
        import importlib
        try:
            lib = importlib.import_module(name)
            if hasattr(lib, 'setup'):
                lib.setup(self)
            self.extensions[name] = lib
        except Exception as e:
            print(f'Failed to load extension {name}: {e}')
            raise e

    def unload_extension(self, name: str):
        """Unload an extension module.
        
        Parameters
        ----------
        name : str
            Module name.
        """
        if name in self.extensions:
            lib = self.extensions[name]
            if hasattr(lib, 'teardown'):
                lib.teardown(self)
            del self.extensions[name]

    def reload_extension(self, name: str):
        """Reload an extension module.
        
        Parameters
        ----------
        name : str
            Module name.
        """
        import importlib
        self.unload_extension(name)
        lib = importlib.import_module(name)
        importlib.reload(lib)
        self.load_extension(name)

    # =========================================
    # Cog Management
    # =========================================

    def add_cog(self, cog: Cog):
        """Add a Cog to the bot.
        
        Parameters
        ----------
        cog : Cog
            The Cog instance.
        """
        cog_name = cog.__class__.__name__
        self.cogs[cog_name] = cog
        
        # Register commands
        for name, member in inspect.getmembers(cog):
            if isinstance(member, Command):
                old_cb = member.callback
                
                # Bind to cog instance
                async def bound_callback(ctx, *args, real_cb=old_cb, cog_ref=cog, **kwargs):
                    return await real_cb(cog_ref, ctx, *args, **kwargs)
                
                bound_callback.__original_func__ = old_cb
                member.callback = bound_callback
                member.cog = cog
                
                # Apply stored cooldown if any
                if hasattr(old_cb, '__command_cooldown__'):
                    member.cooldown = old_cb.__command_cooldown__
                
                self.commands[member.name] = member
                
                # Register aliases
                for alias in member.aliases:
                    self.commands[alias] = member
                
                print(f"Registered cog command: {member.name}")
        
        # Register event listeners
        for name, member in inspect.getmembers(cog):
            if hasattr(member, '__is_listener__') and member.__is_listener__:
                event_name = member.__listener_name__
                
                # Create a bound version
                async def bound_listener(*args, listener=member, cog_ref=cog, **kwargs):
                    return await listener(*args, **kwargs)
                
                self._events[event_name] = bound_listener
                print(f"Registered cog listener: {event_name}")

    def remove_cog(self, name: str) -> Optional[Cog]:
        """Remove a Cog from the bot.
        
        Parameters
        ----------
        name : str
            The Cog class name.
        
        Returns
        -------
        Cog or None
            The removed Cog.
        """
        cog = self.cogs.pop(name, None)
        if cog:
            # Remove commands
            for cmd in cog.get_commands():
                self.commands.pop(cmd.name, None)
                for alias in cmd.aliases:
                    self.commands.pop(alias, None)
        return cog

    def get_cog(self, name: str) -> Optional[Cog]:
        """Get a Cog by name.
        
        Parameters
        ----------
        name : str
            The Cog class name.
        
        Returns
        -------
        Cog or None
        """
        return self.cogs.get(name)

    # =========================================
    # Command Registration
    # =========================================
         
    def command(self, **kwargs):
        """Decorator to register a command on the bot.
        
        Example
        -------
        ```python
        @bot.command(name='ping')
        async def ping(ctx):
            await ctx.send('Pong!')
        ```
        """
        def decorator(func):
            cmd = Command(func, **kwargs)
            
            # Apply stored cooldown
            if hasattr(func, '__command_cooldown__'):
                cmd.cooldown = func.__command_cooldown__
            
            self.commands[cmd.name] = cmd
            
            # Register aliases
            for alias in cmd.aliases:
                self.commands[alias] = cmd
            
            return cmd
        return decorator

    def get_command(self, name: str) -> Optional[Command]:
        """Get a command by name.
        
        Parameters
        ----------
        name : str
            Command name or alias.
        
        Returns
        -------
        Command or None
        """
        return self.commands.get(name)

    # =========================================
    # Command Processing
    # =========================================

    async def on_message(self, message: Message):
        """Handle incoming messages for commands."""
        # Ignore bots
        if message.author and message.author.bot:
            return

        await self.process_commands(message)

    async def process_commands(self, message: Message):
        """Process commands from a message.
        
        Parameters
        ----------
        message : Message
            The message to process.
        """
        content = message.content
        if not content:
            return
        
        # Check prefix
        prefix = self.command_prefix
        if callable(prefix):
            prefix = await prefix(self, message) if asyncio.iscoroutinefunction(prefix) else prefix(self, message)
        
        if not content.startswith(prefix):
            return

        # Parse command using shlex for proper quote handling
        import shlex
        try:
            # We need to handle the case where shlex fails (unclosed quotes)
            # Also, shlex.split might consume the prefix if not careful
            full_content = content[len(prefix):]
            view = shlex.split(full_content)
        except ValueError:
            # Fallback to simple split if quotes are messed up, or raise error
            # User compliant: "mal géré", so let's try to be robust or explicit
            view = content[len(prefix):].split(' ')
            
        if not view:
            return

        invoker = view[0].lower()
        args = view[1:]

        if invoker not in self.commands:
            return

        cmd = self.commands[invoker]
        ctx = Context(
            bot=self, 
            message=message, 
            view=view,
            prefix=prefix,
            command=cmd,
            invoked_with=invoker
        )
        ctx.args = args
        
        try:
            await cmd.invoke(ctx)
        except CommandError as e:
            self.dispatch('command_error', ctx, e)
        except Exception as e:
            self.dispatch('command_error', ctx, e)
            print(f'Command Error: {e}')
            import traceback
            traceback.print_exc()

    async def on_command_error(self, ctx: Context, error: Exception):
        """Default command error handler.
        
        Override this in your bot to customize error handling.
        
        Parameters
        ----------
        ctx : Context
            The command context.
        error : Exception
            The error that occurred.
        """
        if isinstance(error, MissingRequiredArgument):
            await ctx.send(f"❌ Argument manquant : `{error.param.name}`")
        elif isinstance(error, CommandOnCooldown):
            await ctx.send(f"⏰ Commande en cooldown. Réessaie dans {error.retry_after:.1f}s")
        elif isinstance(error, CheckFailure):
            await ctx.send("❌ Tu n'as pas la permission d'utiliser cette commande.")
        elif isinstance(error, BadArgument):
            await ctx.send(f"❌ Argument invalide : {error}")
        else:
            await ctx.send(f"❌ Erreur : {error}")



"""
velmu.permissions
~~~~~~~~~~~~~~~~~

Bitfield flags for Permissions.
"""
from .flags import BaseFlags

class Permissions(BaseFlags):
    """Represents a set of permissions.
    
    The value is a bitfield integer.
    """
    
    CREATE_INSTANT_INVITE = 1 << 0
    KICK_MEMBERS = 1 << 1
    BAN_MEMBERS = 1 << 2
    ADMINISTRATOR = 1 << 3
    MANAGE_CHANNELS = 1 << 4
    MANAGE_GUILD = 1 << 5
    ADD_REACTIONS = 1 << 6
    VIEW_AUDIT_LOG = 1 << 7
    PRIORITY_SPEAKER = 1 << 8
    STREAM = 1 << 9
    VIEW_CHANNEL = 1 << 10
    SEND_MESSAGES = 1 << 11
    SEND_TTS_MESSAGES = 1 << 12
    MANAGE_MESSAGES = 1 << 13
    EMBED_LINKS = 1 << 14
    ATTACH_FILES = 1 << 15
    READ_MESSAGE_HISTORY = 1 << 16
    MENTION_EVERYONE = 1 << 17
    USE_EXTERNAL_EMOJIS = 1 << 18
    VIEW_GUILD_INSIGHTS = 1 << 19
    CONNECT = 1 << 20
    SPEAK = 1 << 21
    MUTE_MEMBERS = 1 << 22
    DEAFEN_MEMBERS = 1 << 23
    MOVE_MEMBERS = 1 << 24
    USE_VAD = 1 << 25
    CHANGE_NICKNAME = 1 << 26
    MANAGE_NICKNAMES = 1 << 27
    MANAGE_ROLES = 1 << 28
    MANAGE_WEBHOOKS = 1 << 29
    MANAGE_EMOJIS_AND_STICKERS = 1 << 30
    USE_APPLICATION_COMMANDS = 1 << 31
    REQUEST_TO_SPEAK = 1 << 32
    MANAGE_EVENTS = 1 << 33
    MANAGE_THREADS = 1 << 34
    CREATE_PUBLIC_THREADS = 1 << 35
    CREATE_PRIVATE_THREADS = 1 << 36
    USE_EXTERNAL_STICKERS = 1 << 37
    SEND_MESSAGES_IN_THREADS = 1 << 38
    USE_EMBEDDED_ACTIVITIES = 1 << 39
    MODERATE_MEMBERS = 1 << 40
    
    @classmethod
    def all(cls):
        """A helper method that sets all permissions to True."""
        all_perms = 0
        for name, value in cls.__dict__.items():
            if isinstance(value, int) and not name.startswith('_'):
                all_perms |= value
        return cls(all_perms)

    @classmethod
    def none(cls):
        """A helper method that sets all permissions to False."""
        return cls(0)
    
    @classmethod
    def general(cls):
        """A helper method that sets permissions generally for public channels."""
        p = cls.none()
        p.view_channel = True
        p.send_messages = True
        p.read_message_history = True
        p.change_nickname = True
        p.use_application_commands = True
        return p
        
    @classmethod
    def from_backend_strings(cls, strings: list):
        """Converts a list of backend permission strings to a Permissions object."""
        value = 0
        # Mapping for discrepancies between Backend and SDK/Discord naming
        mapping = {
            'ADMINISTRATOR': cls.ADMINISTRATOR,
            'MANAGE_SERVER': cls.MANAGE_GUILD,
            'VIEW_CHANNELS': cls.VIEW_CHANNEL,
            'CREATE_INVITES': cls.CREATE_INSTANT_INVITE,
            # Standard mappings
            'KICK_MEMBERS': cls.KICK_MEMBERS,
            'BAN_MEMBERS': cls.BAN_MEMBERS,
            'MANAGE_ROLES': cls.MANAGE_ROLES,
            'MANAGE_CHANNELS': cls.MANAGE_CHANNELS,
            'VIEW_AUDIT_LOG': cls.VIEW_AUDIT_LOG,
            'SEND_MESSAGES': cls.SEND_MESSAGES,
            'READ_MESSAGE_HISTORY': cls.READ_MESSAGE_HISTORY,
            'MENTION_EVERYONE': cls.MENTION_EVERYONE,
            'MANAGE_MESSAGES': cls.MANAGE_MESSAGES,
            'EMBED_LINKS': cls.EMBED_LINKS,
            'ATTACH_FILES': cls.ATTACH_FILES,
            'ADD_REACTIONS': cls.ADD_REACTIONS,
            'CONNECT': cls.CONNECT,
            'SPEAK': cls.SPEAK,
            'MUTE_MEMBERS': cls.MUTE_MEMBERS,
            'DEAFEN_MEMBERS': cls.DEAFEN_MEMBERS,
            'MOVE_MEMBERS': cls.MOVE_MEMBERS
        }
        
        for s in strings:
            if s in mapping:
                value |= mapping[s]
            elif hasattr(cls, s) and isinstance(getattr(cls, s), int):
                value |= getattr(cls, s)
                
        return cls(value)
        
    # --- Properties (Auto-generated in typical libs, but manual here for clarity/simplicity with BaseFlags) ---
    # We can use __getattr__ or just explicit properties. 
    # BaseFlags uses a generator for __repr__ but doesn't auto-create properties.
    # We should add properties for DX (p.administrator = True).
    
    # Actually BaseFlags relies on _set_flag which is internal. 
    # Let's add explicit properties for common ones to be "Pro".
    
    @property
    def administrator(self): return (self.value & self.ADMINISTRATOR) == self.ADMINISTRATOR
    @administrator.setter
    def administrator(self, v): self._set_flag(self.ADMINISTRATOR, v)

    @property
    def manage_guild(self): return (self.value & self.MANAGE_GUILD) == self.MANAGE_GUILD
    @manage_guild.setter
    def manage_guild(self, v): self._set_flag(self.MANAGE_GUILD, v)
    
    @property
    def manage_messages(self): return (self.value & self.MANAGE_MESSAGES) == self.MANAGE_MESSAGES
    @manage_messages.setter
    def manage_messages(self, v): self._set_flag(self.MANAGE_MESSAGES, v)
    
    @property
    def kick_members(self): return (self.value & self.KICK_MEMBERS) == self.KICK_MEMBERS
    @kick_members.setter
    def kick_members(self, v): self._set_flag(self.KICK_MEMBERS, v)

    @property
    def ban_members(self): return (self.value & self.BAN_MEMBERS) == self.BAN_MEMBERS
    @ban_members.setter
    def ban_members(self, v): self._set_flag(self.BAN_MEMBERS, v)
    
    # ... (Adding all is tedious but I'll add the most important ones for now)
    
    @property
    def view_channel(self): return (self.value & self.VIEW_CHANNEL) == self.VIEW_CHANNEL
    @view_channel.setter
    def view_channel(self, v): self._set_flag(self.VIEW_CHANNEL, v)

    @property
    def send_messages(self): return (self.value & self.SEND_MESSAGES) == self.SEND_MESSAGES
    @send_messages.setter
    def send_messages(self, v): self._set_flag(self.SEND_MESSAGES, v)



"""
velmu.errors
~~~~~~~~~~~~

Exception classes for Velmu SDK.
"""

class VelmuException(Exception):
    """Base exception class for all Velmu exceptions."""
    pass

class ClientException(VelmuException):
    """Exception that occurs from the client side."""
    pass

class LoginFailure(ClientException):
    """Exception that occurs when the client fails to log in."""
    pass

class GatewayError(VelmuException):
    """Exception that occurs from the Gateway (WebSocket)."""
    def __init__(self, code=None, message=None, missing_intents=None):
        self.code = code
        self.message = message
        self.missing_intents = missing_intents
        
        fmt = f"Gateway Error {code}: {message}"
        if missing_intents:
            fmt += f" (Missing Intents: {', '.join(missing_intents)})"
        super().__init__(fmt)

class HTTPException(VelmuException):
    """Exception that occurs when an HTTP request fails."""
    def __init__(self, response, message):
        self.response = response
        self.status = response.status
        self.text = message
        if isinstance(message, dict):
            self.text = message.get('message', str(message))
            self.code = message.get('code', 0)
        else:
            self.code = 0
            
        super().__init__(f"{self.status} {self.response.reason}: {self.text}")

class Forbidden(HTTPException):
    """Exception that occurs when the client is forbidden (403)."""
    pass

class NotFound(HTTPException):
    """Exception that occurs when a resource is not found (404)."""
    pass

class VelmuServerError(HTTPException):
    """Exception that occurs when the server returns a 500 error."""
    pass

class RateLimited(HTTPException):
    """Exception that occurs when the client is rate limited (429)."""
    def __init__(self, response, data):
        self.retry_after = data.get('retry_after', 0)
        super().__init__(response, f"Rate limited. Retry in {self.retry_after}s")

# --- Command Errors ---

class CommandError(VelmuException):
    """Base class for command errors."""
    pass

class CheckFailure(CommandError):
    """Exception raised when a check fails."""
    pass

class CommandNotFound(CommandError):
    """Exception raised when a command is not found."""
    pass

class MissingRequiredArgument(CommandError):
    """Exception raised when a required argument is missing."""
    def __init__(self, param):
        self.param = param
        super().__init__(f'{param.name} is a required argument that is missing.')

class BadArgument(CommandError):
    """Exception raised when a bad argument is passed."""
    pass

class CommandOnCooldown(CommandError):
    """Exception raised when the command is on cooldown."""
    def __init__(self, retry_after):
        self.retry_after = retry_after
        super().__init__(f"You are on cooldown. Try again in {retry_after:.2f}s")



"""
velmu.state
~~~~~~~~~~~

Internal state and cache management.
"""
import asyncio
import logging
from typing import Dict, Optional, TYPE_CHECKING, Deque
from collections import deque
from .models import User, Guild, Channel, Message

if TYPE_CHECKING:
    from .client import Client
    from .api import HTTPClient

_log = logging.getLogger(__name__)

class ConnectionState:
    """Manages the internal state of the client (Caches, HTTP, etc)."""
    
    def __init__(self, dispatch, http: 'HTTPClient', max_messages: int = 1000):
        self.dispatch = dispatch
        self.http = http
        self.max_messages = max_messages
        self.user: Optional[User] = None
        
        # Caches
        self._users: Dict[str, User] = {}
        self._guilds: Dict[str, Guild] = {}
        self._channels: Dict[str, Channel] = {}
        self._messages: Deque[Message] = deque(maxlen=max_messages)
        
        # We can add more caches here (roles, members, etc)

    def clear(self):
        self.user = None
        self._users = {}
        self._guilds = {}
        self._channels = {}

    # --- Cache Methods ---
    
    def store_user(self, data):
        user_id = data.get('id')
        user = self._users.get(user_id)
        if user is None:
            user = User(state=self, data=data)
            self._users[user_id] = user
        else:
            user._update(data)
        return user

    def get_user(self, id: str) -> Optional[User]:
        return self._users.get(id)

    def store_message(self, data) -> Message:
        channel_id = data.get('channelId')
        channel = self.get_channel(channel_id)
        
        message = Message(state=self, data=data, channel=channel)
        self._messages.append(message)
        return message

    def _get_message(self, message_id: str) -> Optional[Message]:
        return next((m for m in self._messages if m.id == message_id), None)

    def store_channel(self, data):
        channel_id = data.get('id')
        guild_id = data.get('guildId') or data.get('serverId')
        
        channel = self._channels.get(channel_id)
        if channel is None:
            channel = Channel(state=self, data=data)
            self._channels[channel_id] = channel
        else:
            channel._update(data)
            
        # Link to Guild
        if guild_id:
            guild = self._guilds.get(guild_id)
            if guild:
                guild._channels[channel_id] = channel
                
        return channel

    def store_member(self, guild, data):
        """Stores a Member and links it to the Guild."""
        # data is the full member object: { userId: ..., user: {...}, roles: [], ... }
        user_data = data.get('user')
        if not user_data and 'username' in data:
             # Handle flat user object if necessary (fallback)
             user_data = data
        
        if user_data:
            self.store_user(user_data)
            
        user_id = data.get('userId') or user_data.get('id')
        if not user_id:
            return None
            
        from .models import Member
        member = guild.get_member(user_id)
        if member is None:
            member = Member(state=self, data=data, guild=guild)
            guild._members[user_id] = member
        else:
            member._update(data)
            
        return member

    def get_channel(self, id: str) -> Optional[Channel]:
        return self._channels.get(id)

    def store_guild(self, data):
        guild_id = data.get('id')
        guild = self._guilds.get(guild_id)
        if guild is None:
            guild = Guild(state=self, data=data)
            self._guilds[guild_id] = guild
        else:
            guild._update(data)
        return guild

    def get_guild(self, id: str) -> Optional[Guild]:
        return self._guilds.get(id)

    # --- Event Parsers ---

    def parse_channel_create(self, data) -> Optional[Channel]:
        return self.store_channel(data)

    def parse_channel_update(self, data) -> Optional[Channel]:
        return self.store_channel(data)

    def parse_channel_delete(self, data) -> Optional[Channel]:
        channel_id = data.get('id')
        guild_id = data.get('guildId')
        
        channel = self._channels.pop(channel_id, None)
        
        if guild_id:
            guild = self.get_guild(guild_id)
            if guild and channel_id in guild._channels:
                del guild._channels[channel_id]
                
        return channel

    def parse_guild_member_add(self, data) -> Optional['User']: # Returns Member actually
        guild_id = data.get('guildId')
        guild = self.get_guild(guild_id)
        if guild:
            return self.store_member(guild, data)
        return None

    def parse_guild_member_remove(self, data):
        guild_id = data.get('guildId')
        user_id = data.get('userId') or data.get('user', {}).get('id')
        
        guild = self.get_guild(guild_id)
        if guild and user_id in guild._members:
            return guild._members.pop(user_id)
        return None

    def parse_guild_role_create(self, data):
        guild_id = data.get('guildId')
        role_data = data.get('role')
        
        guild = self.get_guild(guild_id)
        if guild and role_data:
            from .models.guild import Role
            role = Role(self, role_data, guild)
            guild._roles[role.id] = role
            # Also update guild's role list if it's stored separately? 
            # Guild.roles is a dict, so we are good.
            return role
        return None

    def parse_guild_role_update(self, data):
        guild_id = data.get('guildId')
        role_data = data.get('role')
        
        guild = self.get_guild(guild_id)
        if guild and role_data:
            role_id = role_data.get('id')
            role = guild.get_role(role_id)
            if role:
                role._update(role_data)
                return role
        return None

    def parse_guild_role_delete(self, data):
        guild_id = data.get('guildId')
        role_id = data.get('roleId')
        
        guild = self.get_guild(guild_id)
        if guild and role_id in guild._roles:
            return guild._roles.pop(role_id)
        return None

    def parse_presence_update(self, data):
        # Data: { userId, status, guildId, ... }
        user_id = data.get('userId') or data.get('user', {}).get('id')
        status = data.get('status')
        
        # 1. Update Global User
        user = self.get_user(user_id)
        if user:
            setattr(user, 'status', status)
            
        # 2. Fan-out: Update Member in ALL guilds (since they are separate instances)
        # Presence updates might be guild-specific in Discord, but for this bot simplified:
        # We propagate status to all known member instances.
        if user_id:
            for guild in self._guilds.values():
                member = guild.get_member(user_id)
                if member:
                    setattr(member, 'status', status)
                    
        return user

    def parse_guild_roles_position_update(self, data):
        guild_id = data.get('guildId')
        roles_data = data.get('roles', [])
        
        guild = self.get_guild(guild_id)
        if guild:
            updated_roles = []
            for role_datum in roles_data:
                role_id = role_datum.get('id')
                role = guild.get_role(role_id)
                if role:
                    role._update(role_datum)
                    updated_roles.append(role)
            return updated_roles
        return None

    def parse_guild_ban_add(self, data):
        guild_id = data.get('guildId')
        user_data = data.get('user')
        
        guild = self.get_guild(guild_id)
        if guild and user_data:
            # We don't necessarily cache bans, but we return the user and guild info
            # so the client can dispatch 'on_member_ban(guild, user)'
            user = self.store_user(user_data)
            return guild, user
        return None

    def parse_guild_ban_remove(self, data):
        guild_id = data.get('guildId')
        user_data = data.get('user')
        
        guild = self.get_guild(guild_id)
        if guild and user_data:
            user = self.store_user(user_data)
            return guild, user
        return None

    def parse_channel_update_permissions(self, data):
        guild_id = data.get('guildId')
        channel_id = data.get('channelId')
        override = data.get('override')
        
        channel = self.get_channel(channel_id)
        if channel:
            # TODO: Update channel's internal permission overrides list if cached
            # For now return relevant info
            pass
        return channel

    def parse_message_reaction_add(self, data):
        # data: { userId, channelId, messageId, guildId, emoji: { name, id } }
        channel_id = data.get('channelId')
        message_id = data.get('messageId')
        user_id = data.get('userId')
        emoji_data = data.get('emoji')
        
        # Resolve Emoji (simplify to name/str for now if id is null)
        emoji = emoji_data.get('name') if emoji_data else None

        channel = self.get_channel(channel_id)
        if not channel:
            # Partial Channel?
            pass
            
        # Try to find message in cache
        message = self._get_message(message_id)
        if not message and channel:
             # Create partial message?
             # For now, let's try to fetch or just use a partial
             from .models.channel import PartialMessageable
             # Actually we need a Message object for Reaction.message
             # We can construct a partial Message state
             pass

        user = self.get_user(user_id)
        if not user and data.get('guildId'):
             guild = self.get_guild(data.get('guildId'))
             if guild:
                 # Try to fetch member if not in cache?
                 user = guild.get_member(user_id)
        
        if not message:
            # We can't really return a full Reaction object without a Message object
            # But the user might want the event anyway.
            # Let's manufacture a partial message
            from .models.channel import Message
            # Minimal data
            msg_data = {'id': message_id, 'channelId': channel_id, 'guildId': data.get('guildId')}
            message = Message(self, msg_data, channel)
            
        from .models.channel import Reaction
        reaction = Reaction(message, {}, emoji=emoji)
        return reaction, user



"""
velmu.api
~~~~~~~~~

HTTP Client implementation for Velmu.
"""
import aiohttp
import asyncio
import json
import logging
import sys
import weakref
from urllib.parse import quote as _uriquote
from typing import ClassVar, Any, Optional, Dict, Union

from .errors import (
    HTTPException,
    Forbidden,
    NotFound,
    VelmuServerError,
    RateLimited,
    LoginFailure
)
from ._version import __version__
from .config import API_BASE_URL

_log = logging.getLogger(__name__)

class Route:
    """Represents an API Route.
    
    Attributes
    ----------
    method : str
        HTTP Method (GET, POST, PUT, DELETE, PATCH).
    path : str
        API Path template.
    url : str
        Full formatted URL.
    """
    
    BASE: ClassVar[str] = API_BASE_URL

    def __init__(self, method: str, path: str, **parameters: Any):
        self.method = method
        self.path = path
        url = self.BASE + path
        if parameters:
            self.url = url.format(**{k: _uriquote(str(v)) if isinstance(v, str) else v for k, v in parameters.items()})
        else:
            self.url = url
            
        # Bucket hashing for rate limits (simulated for now)
        self.channel_id: Optional[str] = parameters.get('channel_id')
        self.guild_id: Optional[str] = parameters.get('guild_id')

class HTTPClient:
    """Represents an HTTP client sending HTTP requests to the Velmu API."""

    def __init__(self, token: Optional[str] = None):
        self._token: Optional[str] = token
        self._session: Optional[aiohttp.ClientSession] = None
        self._global_over: asyncio.Event = asyncio.Event()
        self._global_over.set()
        self.user_agent: str = f'VelmuBot (https://github.com/velmu-dev/velmu, {__version__}) Python/{sys.version_info[0]}.{sys.version_info[1]} aiohttp/{aiohttp.__version__}'

    async def static_login(self, token: str):
        self._token = token
        self._session = aiohttp.ClientSession()

    async def close(self):
        if self._session:
            await self._session.close()

    async def request(self, route: Route, **kwargs: Any) -> Any:
        if self._session is None:
            self._session = aiohttp.ClientSession()

        headers: Dict[str, str] = {
            'User-Agent': self.user_agent,
            'Content-Type': 'application/json'
        }

        if self._token:
            headers['Authorization'] = f'Bot {self._token}'

        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
            del kwargs['headers']

        if 'json' in kwargs:
            kwargs['data'] = json.dumps(kwargs['json'])
            del kwargs['json']

        reason = kwargs.pop('reason', None)
        if reason:
            headers['X-Audit-Log-Reason'] = _uriquote(reason, safe='/ ')

        # TODO: Real bucket based rate limiting
        # For now, we handle 429s reactively
        
        for attempt in range(5):
            async with self._session.request(route.method, route.url, headers=headers, **kwargs) as response:
                _log.debug(f'{route.method} {route.url} with {kwargs.get("data")} has returned {response.status}')
                
                data = None
                text = await response.text(encoding='utf-8')
                try:
                    content_type = response.headers.get('content-type', '')
                    if 'application/json' in content_type:
                        data = json.loads(text)
                    else:
                        data = text
                except json.JSONDecodeError:
                    _log.error(f"Failed to decode JSON from {route.url}. Response text: {text}")
                    data = text

                # Check for rate limit headers (Standard Discord-like)
                remaining = response.headers.get('X-RateLimit-Remaining')
                if remaining == '0' and response.status != 429:
                    # Pre-emptive wait?
                    pass

                if 300 > response.status >= 200:
                    return data

                if response.status == 429:
                    # Rate Limited
                    retry_after = 0
                    if isinstance(data, dict):
                        retry_after = data.get('retry_after', 0)
                    
                    # Fallback to header
                    if not retry_after:
                         retry_after = float(response.headers.get('Retry-After', 1))

                    _log.warning(f'We are being rate limited. Retrying in {retry_after:.2f} seconds.')
                    
                    # Handle global rate limit
                    is_global = isinstance(data, dict) and data.get('global', False)
                    if is_global:
                        _log.warning('Global rate limit hit. Sleeping.')
                        self._global_over.clear()
                    
                    await asyncio.sleep(retry_after)
                    
                    if is_global:
                        self._global_over.set()
                        
                    continue

                if response.status == 403:
                    raise Forbidden(response, data)
                elif response.status == 404:
                    raise NotFound(response, data)
                elif response.status >= 500:
                    raise VelmuServerError(response, data)
                else:
                    raise HTTPException(response, data)
        
        # If we run out of retries
        raise HTTPException(response, data)

    # --- API Methods ---
    # Organized by Resource
    
    # User
    async def get_me(self):
        """Returns the current bot user."""
        return await self.request(Route('GET', '/users/me'))

    async def get_user(self, user_id: str):
        """Returns a user by ID."""
        return await self.request(Route('GET', f'/users/{user_id}'))

    async def create_dm(self, recipient_id: str):
        """Creates a DM channel with a user."""
        # VELMU API uses /conversations for DMs
        return await self.request(Route('POST', '/conversations'), json={'targetUserId': recipient_id})

    # Guild
    async def get_guild(self, guild_id: str):
        """Returns a guild by ID."""
        return await self.request(Route('GET', f'/servers/{guild_id}'))
        
    async def get_guild_channels(self, guild_id: str):
        """Returns all channels in a guild."""
        return await self.request(Route('GET', f'/servers/{guild_id}/channels'))

    async def get_member(self, guild_id: str, user_id: str):
        """Returns a member of a guild."""
        return await self.request(Route('GET', f'/members/{guild_id}/{user_id}'))

    # Channel
    async def get_channel(self, channel_id: str):
        """Returns a channel by ID."""
        return await self.request(Route('GET', f'/channels/{channel_id}'))

    # Messages
    async def send_message(self, channel_id: str, content: Optional[str], embed: Optional[Dict], reply_to: Optional[str] = None):
        """Sends a message to a channel."""
        payload = {}
        if content: payload['content'] = content
        if embed: payload['embed'] = embed
        if reply_to: payload['replyToId'] = reply_to
        
        return await self.request(Route('POST', f'/channels/{channel_id}/messages'), json=payload)

    async def send_conversation_message(self, conversation_id: str, content: Optional[str], embed: Optional[Dict], reply_to: Optional[str] = None):
        """Sends a message to a DM conversation."""
        payload = {'conversationId': conversation_id}
        if content: payload['content'] = content
        if embed: payload['embed'] = embed
        if reply_to: payload['replyToId'] = reply_to
        
        return await self.request(Route('POST', '/messages'), json=payload)

    async def get_messages(self, channel_id: str, limit: int = 50, before: Optional[str] = None, after: Optional[str] = None):
        """Returns messages from a channel."""
        params = {'limit': limit}
        if before:
            params['before'] = before
        if after:
            params['after'] = after
            
        return await self.request(Route('GET', f'/channels/{channel_id}/messages'), params=params)

    async def get_message(self, channel_id: str, message_id: str):
        """Returns a specific message."""
        return await self.request(Route('GET', f'/messages/{message_id}'))

    async def delete_message(self, message_id: str):
        """Deletes a message."""
        return await self.request(Route('DELETE', f'/messages/{message_id}'))

    async def edit_message(self, message_id: str, content: Optional[str], embed: Optional[Dict]):
        """Edits a message."""
        payload = {}
        if content is not None: payload['content'] = content
        if embed is not None: payload['embed'] = embed
        return await self.request(Route('PUT', f'/messages/{message_id}'), json=payload)
    
    async def add_reaction(self, message_id: str, emoji: str):
         """Adds a reaction to a message."""
         return await self.request(Route('POST', f'/messages/{message_id}/reactions'), json={'emoji': emoji})

    async def remove_reaction(self, message_id: str, emoji: str, user_id: Optional[str] = None):
        """Removes a reaction from a message.
        
        If user_id is None, removes the bot's own reaction.
        """
        params = {}
        if user_id: params['userId'] = user_id
        return await self.request(Route('DELETE', f'/messages/{message_id}/reactions/{emoji}'), params=params)

    # Pins
    async def pin_message(self, message_id: str):
        """Pins a message in its channel."""
        return await self.request(Route('PUT', f'/messages/{message_id}/pin'))

    async def unpin_message(self, message_id: str):
        """Unpins a message."""
        return await self.request(Route('DELETE', f'/messages/{message_id}/pin'))

    async def get_channel_pins(self, channel_id: str):
        """Returns all pinned messages in a channel."""
        return await self.request(Route('GET', f'/channels/{channel_id}/pins'))

    async def get_conversation_pins(self, conversation_id: str):
        """Returns all pinned messages in a conversation."""
        return await self.request(Route('GET', f'/conversations/{conversation_id}/pins'))

    # Invites
    async def get_invite(self, code: str):
        """Returns invite info."""
        return await self.request(Route('GET', f'/invites/{code}'))

    # Channel/Category Management
    async def create_channel(self, guild_id: str, name: str, type: int, category_id: str = None, permission_overwrites = None):
        """Creates a channel."""
        payload = {
            'serverId': guild_id,
            'name': name, 
            'type': type
        }
        if category_id:
            payload['categoryId'] = category_id
        if permission_overwrites:
            payload['permissionOverwrites'] = permission_overwrites
            
        return await self.request(Route('POST', '/channels'), json=payload)

    async def delete_channel(self, channel_id: str):
        """Deletes a channel."""
        return await self.request(Route('DELETE', f'/channels/{channel_id}'))

    async def create_category(self, guild_id: str, name: str):
        """Creates a new category in a guild."""
        payload = {
            'serverId': guild_id,
            'name': name
        }
        return await self.request(Route('POST', '/categories'), json=payload)

    async def delete_category(self, category_id: str):
        """Deletes a category."""
        return await self.request(Route('DELETE', f'/categories/{category_id}'))

    async def edit_channel(self, channel_id: str, **kwargs):
        """Edits channel settings."""
        return await self.request(Route('PUT', f'/channels/{channel_id}'), json=kwargs)

    async def edit_guild(self, guild_id: str, **kwargs):
        """Edits guild settings."""
        return await self.request(Route('PUT', f'/servers/{guild_id}'), json=kwargs)

    async def edit_member(self, guild_id: str, user_id: str, **kwargs):
        """Edits a member (e.g. nickname)."""
        return await self.request(Route('PUT', f'/members/{guild_id}/update/{user_id}'), json=kwargs)

    # Moderation
    async def kick_member(self, guild_id: str, user_id: str):
        """Kicks a member from the guild."""
        return await self.request(Route('DELETE', f'/members/{guild_id}/kick/{user_id}'))

    async def ban_member(self, guild_id: str, user_id: str, reason: Optional[str] = None, expires_at: Optional[str] = None):
        """Bans a member from the guild."""
        payload = {}
        if reason: payload['reason'] = reason
        if expires_at: payload['expiresAt'] = expires_at
        return await self.request(Route('POST', f'/servers/{guild_id}/bans/{user_id}'), json=payload)

    async def unban_member(self, guild_id: str, user_id: str):
        """Unbans a member."""
        return await self.request(Route('DELETE', f'/servers/{guild_id}/bans/{user_id}'))
    
    async def get_bans(self, guild_id: str):
        """Returns a list of bans for a guild."""
        return await self.request(Route('GET', f'/servers/{guild_id}/bans'))

    # Roles
    async def get_roles(self, guild_id: str):
        """Returns all roles in a guild."""
        return await self.request(Route('GET', f'/servers/{guild_id}/roles'))

    async def add_role(self, guild_id: str, user_id: str, role_id: str):
        """Assigns a role to a member."""
        return await self.request(Route('PUT', f'/members/{guild_id}/{user_id}/roles/{role_id}'))

    async def remove_role(self, guild_id: str, user_id: str, role_id: str):
        """Removes a role from a member."""
        return await self.request(Route('DELETE', f'/members/{guild_id}/{user_id}/roles/{role_id}'))

    async def create_role(self, guild_id: str, name: Optional[str] = None, color: Optional[str] = None, permissions: int = 0):
        """Creates a new role.
        
        Note: Backend might require a separate update call for properties if not supported in create.
        """
        # Current backend implementation doesn't support body in create.
        # We create then update.
        role = await self.request(Route('POST', f'/servers/{guild_id}/roles'))
        if name or color or permissions:
             role_id = role.get('id')
             if role_id:
                 payload = {}
                 if name: payload['name'] = name
                 if color: payload['color'] = color
                 if permissions: payload['permissions'] = permissions
                 return await self.edit_role(guild_id, role_id, **payload)
        return role

    async def edit_role(self, guild_id: str, role_id: str, **kwargs):
        """Edits a role's permissions or settings."""
        return await self.request(Route('PUT', f'/servers/{guild_id}/roles/{role_id}'), json=kwargs)

    async def delete_role(self, guild_id: str, role_id: str):
        """Deletes a role."""
        return await self.request(Route('DELETE', f'/servers/{guild_id}/roles/{role_id}'))
    
    async def edit_role_positions(self, guild_id: str, roles: list):
        """Updates role hierarchy positions."""
        return await self.request(Route('PUT', f'/servers/{guild_id}/roles/positions'), json={'roles': roles})



from typing import List, Dict, Union, Optional
import datetime

class Embed:
    """Represents a rich embed object."""
    
    def __init__(self, **kwargs):
        self.title = kwargs.get('title')
        self.type = kwargs.get('type', 'rich')
        self.description = kwargs.get('description')
        self.url = kwargs.get('url')
        self.timestamp = kwargs.get('timestamp')
        self.color = kwargs.get('color')
        
        self.footer = kwargs.get('footer')
        self.image = kwargs.get('image')
        self.thumbnail = kwargs.get('thumbnail')
        self.video = kwargs.get('video')
        self.provider = kwargs.get('provider')
        self.author = kwargs.get('author')
        self.fields = kwargs.get('fields', [])
        
    def to_dict(self) -> Dict:
        """Converts the embed to a dictionary."""
        result = {'type': self.type}
        
        if self.title:
            result['title'] = self.title
        if self.description:
            result['description'] = self.description
        if self.url:
            result['url'] = self.url
        if self.timestamp:
            if isinstance(self.timestamp, datetime.datetime):
                result['timestamp'] = self.timestamp.isoformat()
            else:
                result['timestamp'] = self.timestamp
        if self.color:
            result['color'] = self.color
            
        if self.footer:
            result['footer'] = self.footer
        if self.image:
            result['image'] = self.image
        if self.thumbnail:
            result['thumbnail'] = self.thumbnail
        if self.author:
            result['author'] = self.author
        if self.fields:
            result['fields'] = self.fields
            
        return result

    def set_footer(self, text: str, icon_url: Optional[str] = None):
        """Sets the footer of the embed."""
        self.footer = {'text': text}
        if icon_url:
            self.footer['icon_url'] = icon_url
        return self

    def set_image(self, url: str):
        """Sets the image of the embed."""
        self.image = {'url': url}
        return self

    def set_thumbnail(self, url: str):
        """Sets the thumbnail of the embed."""
        self.thumbnail = {'url': url}
        return self

    def set_author(self, name: str, url: Optional[str] = None, icon_url: Optional[str] = None):
        """Sets the author of the embed."""
        self.author = {'name': name}
        if url:
            self.author['url'] = url
        if icon_url:
            self.author['icon_url'] = icon_url
        return self

    def add_field(self, name: str, value: str, inline: bool = True):
        """Adds a field to the embed."""
        field = {
            'name': name,
            'value': value,
            'inline': inline
        }
        self.fields.append(field)
        return self



# velmu.ext.tasks
import asyncio
import inspect
import logging
import sys
import traceback
from typing import Any, Callable, Coroutine, Optional, Union, TypeVar

_log = logging.getLogger(__name__)

LF = TypeVar('LF', bound=Callable[..., Coroutine[Any, Any, Any]])

class Loop:
    """
    A background task loop.
    """
    def __init__(
        self, 
        coro: LF, 
        seconds: float = 0, 
        minutes: float = 0, 
        hours: float = 0, 
        count: Optional[int] = None, 
        reconnect: bool = True, 
        loop: Optional[asyncio.AbstractEventLoop] = None
    ):
        self.coro = coro
        self.seconds = seconds
        self.minutes = minutes
        self.hours = hours
        self.count = count
        self.reconnect = reconnect
        self._loop = loop or asyncio.get_event_loop()
        
        self._task: Optional[asyncio.Task] = None
        self._injected_loop = None
        self._valid_exception = (OSError, asyncio.TimeoutError) # Exceptions to ignore handling if reconnect=True?
        
        self._before_loop = None
        self._after_loop = None
        self._error = None

        self._stop_next_iteration = False
        self._current_loop = 0
        self._next_iteration = None
        
        if self.seconds == 0 and self.minutes == 0 and self.hours == 0:
            raise ValueError("You must supply a time interval")

    async def _loop_logic(self, *args, **kwargs):
        try:
            if self._before_loop:
                await self._before_loop(*args, **kwargs)

            self._next_iteration = self._loop.time()
            # Loop
            while True:
                if self._stop_next_iteration:
                    return

                # Calculate Sleep
                now = self._loop.time()
                if self._next_iteration is None: 
                    self._next_iteration = now
                    
                if self._next_iteration > now:
                    await asyncio.sleep(self._next_iteration - now)
                
                # Execute
                try:
                    await self.coro(*args, **kwargs)
                    self._current_loop += 1
                    
                    if self.count is not None and self._current_loop >= self.count:
                        break
                except Exception as exc:
                    # Handle Error
                    if self._error:
                        await self._error(exc)
                    else:
                        print(f"Unhandled exception in task {self.coro.__name__}:", file=sys.stderr)
                        traceback.print_exc()
                        
                    if not self.reconnect:
                        raise exc

                # Schedule Next
                interval = self.seconds + (self.minutes * 60.0) + (self.hours * 3600.0)
                self._next_iteration = now + interval
                
        except asyncio.CancelledError:
            self._stop_next_iteration = False
            raise
        except Exception as e:
            # Fatal error
            print(f"Internal loop error: {e}")
        finally:
             if self._after_loop:
                await self._after_loop(*args, **kwargs)
             self._task = None

    def start(self, *args, **kwargs):
        """Starts the loop."""
        if self._task and not self._task.done():
            raise RuntimeError("Task is already running")
        
        self._task = self._loop.create_task(self._loop_logic(*args, **kwargs))
        return self._task

    def stop(self):
        """Stops the loop gracefully after current iteration."""
        if self._task and not self._task.done():
            self._stop_next_iteration = True

    def cancel(self):
        """Cancels the loop immediately."""
        if self._task and not self._task.done():
            self._task.cancel()

    def restart(self, *args, **kwargs):
        """Restarts the loop."""
        self.cancel()
        self.start(*args, **kwargs)
        
    def before_loop(self, coro: LF) -> LF:
        self._before_loop = coro
        return coro

    def after_loop(self, coro: LF) -> LF:
        self._after_loop = coro
        return coro
        
    def error(self, coro: Callable[[Exception], Coroutine[Any, Any, Any]]):
        self._error = coro
        return coro


def loop(
    seconds: float = 0, 
    minutes: float = 0, 
    hours: float = 0, 
    count: Optional[int] = None, 
    reconnect: bool = True, 
    loop: Optional[asyncio.AbstractEventLoop] = None
):
    """
    Decorator to create a background task loop.
    """
    def decorator(func: LF) -> Loop:
        return Loop(func, seconds=seconds, minutes=minutes, hours=hours, count=count, reconnect=reconnect, loop=loop)
    return decorator



"""
velmu.models.user
~~~~~~~~~~~~~~~~~

User and Member models.
"""
from typing import Optional, List, TYPE_CHECKING, Union
from ..abc import Snowflake, Messageable
from ..utils import parse_time

if TYPE_CHECKING:
    from ..state import ConnectionState
    from .guild import Guild

class User(Snowflake, Messageable):
    """Represents a Velmu User."""
    
    __slots__ = ('_state', 'id', 'username', 'discriminator', 'avatar_url', 'bot', 'global_name', 'status')
    
    def __init__(self, state: 'ConnectionState', data: dict):
        self._state = state
        self.id = data.get('id')
        self._update(data)

    def _update(self, data: dict):
        self.username = data.get('username')
        self.discriminator = data.get('discriminator')
        self.avatar_url = data.get('avatarUrl')
        self.bot = data.get('bot', False)
        self.global_name = data.get('globalName')

    @property
    def mention(self) -> str:
        return f'<@{self.id}>'
        
    @property
    def display_name(self) -> str:
        return self.global_name or self.username

    async def _get_channel(self):
        return await self.create_dm()

    async def create_dm(self):
        """Creates a DM channel with this user."""
        data = await self._state.http.create_dm(self.id)
        return self._state.store_channel(data)

    def __repr__(self):
        return f'<User id={self.id} name={self.username!r} bot={self.bot}>'

class Member(User):
    """Represents a Member of a Guild."""
    
    __slots__ = ('guild', 'joined_at', 'nick', 'roles')
    
    def __init__(self, state: 'ConnectionState', data: dict, guild: 'Guild'):
        self.guild = guild
        # data usually contains a 'user' object
        user_data = data.get('user', {})
        super().__init__(state, user_data)
        if self.id is None:
            self.id = data.get('userId')
        self._update_member(data)
        
    def _update_member(self, data: dict):
        self.joined_at = parse_time(data.get('joinedAt'))
        self.nick = data.get('nick')
        # Handle cases where roles is a list of dicts (full objects) or strings (IDs)
        raw_roles = data.get('roles', [])
        processed_roles = []
        for r in raw_roles:
            role_id = r if isinstance(r, str) else r.get('id')
            role = self.guild.get_role(role_id)
            if role:
                processed_roles.append(role)
        self.roles = processed_roles

    @property
    def display_name(self) -> str:
        return self.nick or super().display_name

    async def edit(self, *, nick: Optional[str] = None):
        """Edits the member.
        
        Parameters
        -----------
        nick: Optional[str]
            The new nickname for the member.
        """
        payload = {}
        if nick is not None:
             payload['nickname'] = nick
             
        if payload:
            await self._state.http.edit_member(self.guild.id, self.id, **payload)

    async def kick(self, *, reason: Optional[str] = None):
        """Kicks this member from the guild."""
        await self.guild.kick(self, reason=reason)

    async def ban(self, *, reason: Optional[str] = None):
        """Bans this member from the guild."""
        await self.guild.ban(self, reason=reason)

    async def add_roles(self, *roles: Union['Snowflake', str], reason: Optional[str] = None):
        """Adds roles to this member.
        
        Parameters
        -----------
        *roles: Union[Role, str]
            The roles to add. Can be Role objects or role IDs.
        reason: Optional[str]
            The reason for adding these roles.
        """
        for role in roles:
            role_id = role.id if hasattr(role, 'id') else role
            await self._state.http.add_role(self.guild.id, self.id, role_id)

    async def remove_roles(self, *roles: Union['Snowflake', str], reason: Optional[str] = None):
        """Removes roles from this member."""
        for role in roles:
            role_id = role.id if hasattr(role, 'id') else role
            await self._state.http.remove_role(self.guild.id, self.id, role_id)

    def __repr__(self):
        return f'<Member id={self.id} name={self.username!r} guild={self.guild.id!r}>'



"""
velmu.models.channel
~~~~~~~~~~~~~~~~~~~~

Channel and Message models.
"""
from typing import Optional, List, Union, TYPE_CHECKING
from ..abc import Snowflake, Messageable
from ..utils import parse_time
from .user import User, Member
import asyncio

if TYPE_CHECKING:
    from ..state import ConnectionState
    from .guild import Guild

class ChannelType:
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4

class HistoryIterator:
    """Async iterator for channel history."""
    
    def __init__(self, channel, limit=100, before=None, after=None):
        self.channel = channel
        self.limit = limit
        self.before = before
        self.after = after
        
        self._messages = asyncio.Queue()
        self._retrieve_messages = True
    
    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._messages.empty():
            if not self._retrieve_messages:
                raise StopAsyncIteration
                
            await self._fill_messages()
            
            if self._messages.empty():
                raise StopAsyncIteration
        
        return await self._messages.get()

    async def _fill_messages(self):
        if self.limit <= 0:
            self._retrieve_messages = False
            return

        fetch_limit = min(self.limit, 100) 
        
        data = await self.channel._state.http.get_messages(
            self.channel.id, 
            limit=fetch_limit, 
            before=self.before,
            after=self.after
        )
        
        
        # Handle pagination object (items + cursor)
        if isinstance(data, dict) and 'items' in data:
            data = data['items']

        if not data:
            self._retrieve_messages = False
            return

        if not isinstance(data, list):
            # API returned something unexpected (error string?)
            print(f"History Error: Expected list, got {type(data)}: {data}")
            self._retrieve_messages = False
            return
            
        found = 0
        for msg_data in data:
            self._messages.put_nowait(Message(self.channel._state, msg_data, self.channel))
            found += 1
            
            # Update pointers (Legacy: usually traverse backwards)
            self.before = msg_data['id'] 
            
        self.limit -= found
        if self.limit <= 0 or found < fetch_limit:
            self._retrieve_messages = False

class Channel(Snowflake, Messageable):
    """Represents a Guild or DM Channel."""
    
    __slots__ = ('_state', 'id', 'name', 'type', 'guild_id', 'position', 'permission_overwrites', 'category_id', 'topic', 'last_message_id', 'nsfw')
    
    def __init__(self, state: 'ConnectionState', data: dict, guild: Optional['Guild'] = None):
        self._state = state
        self.id = data.get('id')
        self._update(data, guild)
        
    def _update(self, data: dict, guild: Optional['Guild'] = None):
        self.name = data.get('name')
        self.type = data.get('type')
        self.guild_id = data.get('guildId') or data.get('serverId')
        if not self.guild_id and guild:
            self.guild_id = guild.id
        self.position = data.get('position')
        self.topic = data.get('topic')
        self.category_id = data.get('categoryId')
        self.nsfw = data.get('nsfw', False)
        
    @property
    def guild(self) -> Optional['Guild']:
        if self.guild_id:
            return self._state.get_guild(self.guild_id)
        return None
        
    @property
    def mention(self) -> str:
        return f'<#{self.id}>'

    async def _get_channel(self):
        return self

    async def send(self, content=None, *, embed=None, **kwargs):
        """Sends a message to the channel."""
        # Convert embed object to dict if necessary
        if embed and hasattr(embed, 'to_dict'):
            embed = embed.to_dict()
            
        if self.guild_id or getattr(self, 'serverId', None):
            data = await self._state.http.send_message(self.id, content, embed)
        else:
            # Assume Conversation/DM
            data = await self._state.http.send_conversation_message(self.id, content, embed)
            
        return Message(self._state, data, self)

    async def fetch_message(self, message_id: str) -> 'Message':
        data = await self._state.http.get_message(self.id, message_id)
        return Message(self._state, data, self)

    async def edit(self, *, name: str = None, topic: str = None, position: int = None, nsfw: bool = None, category: Optional[Union['Channel', str]] = None, overwrites: list = None):
        """Edits the channel."""
        payload = {}
        if name is not None: payload['name'] = name
        if topic is not None: payload['topic'] = topic
        if position is not None: payload['position'] = position
        if nsfw is not None: payload['nsfw'] = nsfw
        
        if category is not None:
             if isinstance(category, str):
                 payload['categoryId'] = category
             elif hasattr(category, 'id'):
                 payload['categoryId'] = category.id
             else:
                 payload['categoryId'] = None # Remove category

        if overwrites is not None:
            payload['overwrites'] = overwrites
            
        await self._state.http.edit_channel(self.id, **payload)

    def history(self, *, limit: int = 100, before=None, after=None) -> HistoryIterator:
        """Returns an AsyncIterator for channel history."""
        return HistoryIterator(self, limit=limit, before=before, after=after)

    async def delete(self):
        """Deletes the channel or category."""
        if self.type == ChannelType.GUILD_CATEGORY:
            await self._state.http.delete_category(self.id)
        else:
            await self._state.http.delete_channel(self.id)

    async def pins(self) -> List['Message']:
        """Returns a list of pinned messages in the channel."""
        if self.guild_id or getattr(self, 'serverId', None):
             data = await self._state.http.get_channel_pins(self.id)
        else:
             data = await self._state.http.get_conversation_pins(self.id)
        
        return [Message(self._state, m, self) for m in data]

class PartialMessageable(Snowflake, Messageable):
    """Represents a partial messageable entity (e.g. just a channel ID)."""
    
    __slots__ = ('_state', 'id', 'guild_id', 'type')
    
    def __init__(self, state, id: str, guild_id: Optional[str] = None, type: Optional[int] = None):
        self._state = state
        self.id = id
        self.guild_id = guild_id
        self.type = type
        
    async def _get_channel(self):
        return self
        
    async def edit(self, **kwargs):
        """Edits the channel (Partial Support)."""
        await self._state.http.edit_channel(self.id, **kwargs)

    async def send(self, content=None, *, embed=None, **kwargs):
        """Sends a message to the channel (Partial Support)."""
        if embed and hasattr(embed, 'to_dict'):
            embed = embed.to_dict()
        # We don't know if it's a DM or Guild channel easily without type/guild_id
        # But for 'send_message' we assume standard route
        await self._state.http.send_message(self.id, content, embed)

    @property
    def guild(self) -> Optional['Guild']:
        if self.guild_id:
            return self._state.get_guild(self.guild_id)
        return None

class Message(Snowflake):
    """Represents a Message in a Channel."""
    
    __slots__ = ('_state', 'id', 'content', 'channel_id', 'author', 'embeds', 'timestamp', 'edited_timestamp', 'tts', 'mention_everyone', 'mentions', 'mention_roles', 'attachments', 'pinned', 'webhook_id', 'type', 'activity', 'application', 'reference', 'flags', '_guild_id', 'deleter_id')
    
    def __init__(self, state: 'ConnectionState', data: dict, channel: Optional[Channel] = None):
        self._state = state
        self.id = data.get('id')
        self.channel_id = data.get('channelId')
        self._guild_id = data.get('guildId') or data.get('serverId') # Store this!
        self._update(data, channel)
        
    def _update(self, data: dict, channel: Optional[Channel] = None):
        self.content = data.get('content')
        self.timestamp = parse_time(data.get('timestamp'))
        self.pinned = data.get('isPinned', False)
        self.embeds = data.get('embeds', [])
        
        # Author
        user_data = data.get('author')
        if user_data:
            # 1. Try to resolve as Member if in a Guild
            guild_id = data.get('guildId') or data.get('serverId')
            member_found = None
            if guild_id:
                guild = self._state.get_guild(guild_id)
                if guild:
                    # Try cache
                    member_found = guild.get_member(user_data.get('id'))
                    
            if member_found:
                self.author = member_found
            else:
                self.author = User(self._state, user_data)
        else:
            # Fallback: Try to resolve from cache using userId
            user_id = data.get('userId')
            if user_id:
                # Try global cache (User)
                self.author = self._state.get_user(user_id)
                if not self.author:
                     # Create a minimal User object if not in cache (Critical Fix)
                     # We assume it's a valid user ID, just not cached.
                     self.author = User(self._state, {'id': user_id, 'username': 'Unknown User', 'discriminator': '0000'})

                # Try guild cache (Member) if available
                guild_id = data.get('guildId')
                if guild_id:
                    guild = self._state.get_guild(guild_id)
                    if guild:
                       member = guild.get_member(user_id)
                       if member: self.author = member
            else:
                self.author = None
        
    @property
    def channel(self) -> Union[Channel, 'PartialMessageable']:
        print(f"[SDK DEBUG] Accessing Message.channel property for msg {self.id}")
        channel = self._state.get_channel(self.channel_id)
        if channel:
            print(f"[SDK DEBUG] Found cached channel {channel.name} ({channel.id})")
            return channel
        # Fallback to PartialMessageable if channel is not in cache (e.g. DMs or Cold Cache)
        # We try to infer if it's a DM or Guild Channel?
        # For edits, ID is enough.
        print(f"[SDK DEBUG] Channel {self.channel_id} not in cache. Creating PartialMessageable.")
        
        # Try to resolve guild_id from Message data or Author
        guild_id = getattr(self, '_guild_id', None)
        return PartialMessageable(self._state, self.channel_id, guild_id=guild_id)

    @property
    def guild(self) -> Optional['Guild']:
        channel = self.channel
        return channel.guild if channel else None

    async def delete(self):
        await self._state.http.delete_message(self.id)
        
    async def edit(self, content=None, embed=None):
        if embed and hasattr(embed, 'to_dict'):
            embed = embed.to_dict()
        return await self._state.http.edit_message(self.id, content, embed)
        
    async def reply(self, content=None, embed=None):
        return await self.channel.send(content, embed=embed)

    async def add_reaction(self, emoji: str):
        """Adds a reaction to the message."""
        return await self._state.http.add_reaction(self.id, emoji)

    async def remove_reaction(self, emoji: str, user_id: str = None):
        """Removes a reaction from the message."""
        return await self._state.http.remove_reaction(self.id, emoji, user_id)

    async def pin(self):
        """Pins the message."""
        await self._state.http.pin_message(self.id)
        self.pinned = True

    async def unpin(self):
        """Unpins the message."""
        await self._state.http.unpin_message(self.id)
        self.pinned = False




class Reaction:
    """Represents a reaction to a message."""
    __slots__ = ('message', 'count', 'emoji', 'me')

    def __init__(self, message, data, emoji=None):
        self.message = message
        self.emoji = emoji or data.get('emoji')
        self.count = data.get('count', 1)
        self.me = data.get('me', False)

    def is_custom_emoji(self):
        return not isinstance(self.emoji, str)

    def __repr__(self):
        return f'<Reaction emoji={self.emoji!r} count={self.count} message={self.message!r}>'



import os

# Base URLs
# Default to Production, but allow overrides via env vars for local dev
API_BASE_URL = os.getenv('VELMU_API_URL', 'https://velmuv2-backend.onrender.com/api')
GATEWAY_URL = os.getenv('VELMU_WS_URL', 'https://velmuv2-backend.onrender.com')



"""
velmu.models.guild
~~~~~~~~~~~~~~~~~~

Guild and Role models.
"""
from typing import Optional, List, Dict, TYPE_CHECKING, Union
from ..abc import Snowflake
from ..utils import parse_time
from .user import Member

if TYPE_CHECKING:
    from ..state import ConnectionState
    from .channel import Channel

class Role(Snowflake):
    """Represents a Guild Role."""
    
    __slots__ = ('_state', 'guild', 'id', 'name', 'color', 'hoist', 'position', 'permissions', 'managed', 'mentionable')
    
    def __init__(self, state: 'ConnectionState', data: dict, guild: 'Guild'):
        self._state = state
        self.guild = guild
        self.id = data.get('id')
        self._update(data)
        
    def _update(self, data: dict):
        self.name = data.get('name')
        self.color = data.get('color')
        self.hoist = data.get('hoist', False)
        self.position = data.get('position', 0)
        
        # Permissions conversion
        perm_value = data.get('permissions', 0)
        
        from ..permissions import Permissions
        
        # Handle list of strings (Velmu Backend format)
        if isinstance(perm_value, list):
            self.permissions = Permissions.from_backend_strings(perm_value)
        # Handle string integer (API sometimes sends strings for large bitfields)
        elif isinstance(perm_value, str):
            self.permissions = Permissions(int(perm_value))
        else:
            self.permissions = Permissions(perm_value)
        
        self.managed = data.get('managed', False)
        self.mentionable = data.get('mentionable', False)

    @property
    def mention(self) -> str:
        return f'<@&{self.id}>'

    async def edit(self, *, name: str = None, color: Union[int, str] = None, permissions: int = None, position: int = None, hoist: bool = None, mentionable: bool = None):
        """Edits the role."""
        payload = {}
        if name is not None: payload['name'] = name
        if color is not None:
             if isinstance(color, int):
                 payload['color'] = f'#{color:06x}'
             else:
                 payload['color'] = color
                 
        if permissions is not None: payload['permissions'] = permissions
        if position is not None: payload['position'] = position
        if hoist is not None: payload['hoist'] = hoist
        if mentionable is not None: payload['mentionable'] = mentionable
        
        await self._state.http.edit_role(self.guild.id, self.id, **payload)

    async def delete(self):
        """Deletes the role."""
        await self._state.http.delete_role(self.guild.id, self.id)

class Guild(Snowflake):
    """Represents a Velmu Guild (Server)."""
    
    __slots__ = (
        '_state', 'id', 'name', 'icon', 'owner_id', 'region', 
        'afk_channel_id', 'afk_timeout', 'member_count', 
        'verification_level', 'default_message_notifications',
        'explicit_content_filter', 'emojis', 'features',
        'mfa_level', 'application_id', 'system_channel_id',
        'rules_channel_id', 'joined_at', 'large', 'unavailable',
        'member_count', '_channels', '_members', '_roles'
    )
    
    def __init__(self, state: 'ConnectionState', data: dict):
        self._state = state
        self.id = data.get('id')
        self._channels: Dict[str, 'Channel'] = {}
        self._members: Dict[str, Member] = {}
        self._roles: Dict[str, Role] = {}
        self._update(data)
        
    def _update(self, data: dict):
        self.name = data.get('name')
        self.icon = data.get('icon')
        self.owner_id = data.get('ownerId')
        self.member_count = data.get('memberCount', 0)
    
        # Roles
        if 'roles' in data:
            self._roles.clear() 
            for r_data in data['roles']:
                role = Role(self._state, r_data, self)
                self._roles[role.id] = role

    @property
    def icon_url(self) -> Optional[str]:
        if self.icon:
             if self.icon.startswith('http'):
                 return self.icon
             return None 
        return None
        
    @property
    def roles(self) -> List[Role]:
        """Returns a list of roles in the guild."""
        return list(self._roles.values())

    @property
    def channels(self) -> List['Channel']:
        return list(self._channels.values())

    @property
    def members(self) -> List[Member]:
        return list(self._members.values())

    @property
    def text_channels(self) -> List['Channel']:
        """Returns a list of text channels in the guild."""
        return [c for c in self.channels if str(c.type) == 'TEXT' or c.type == 0]

    @property
    def voice_channels(self) -> List['Channel']:
        """Returns a list of voice channels in the guild."""
        return [c for c in self.channels if str(c.type) == 'AUDIO' or str(c.type) == 'VOICE' or c.type == 2]

    @property
    def owner(self) -> Optional[Member]:
        return self.get_member(self.owner_id)

    def get_member(self, user_id: str) -> Optional[Member]:
        return self._members.get(user_id)

    async def fetch_member(self, user_id: str) -> Member:
        data = await self._state.http.get_member(self.id, user_id)
        member = Member(self._state, data, self)
        self._members[user_id] = member
        return member

    def get_role(self, role_id: str) -> Optional[Role]:
        return self._roles.get(role_id)
        
    def get_channel(self, channel_id: str) -> Optional['Channel']:
        return self._channels.get(channel_id)

    async def edit(self, *, name: str = None, icon: Optional[bytes] = None):
        """Edits the guild."""
        payload = {}
        if name is not None:
             payload['name'] = name
        
        if payload:
            await self._state.http.edit_guild(self.id, **payload)

    async def create_text_channel(self, name: str, category: Optional[Union['Channel', str]] = None, overwrites: list = None) -> 'Channel':
        """Creates a text channel in the guild."""
        from .channel import Channel, ChannelType
        
        category_id = None
        if category:
            if isinstance(category, str):
                category_id = category
            elif hasattr(category, 'id'):
                category_id = category.id
                
        data = await self._state.http.create_channel(self.id, name, ChannelType.GUILD_TEXT, category_id, overwrites)
        channel = Channel(self._state, data, self)
        self._channels[channel.id] = channel
        return channel

    async def create_voice_channel(self, name: str, category: Optional[Union['Channel', str]] = None) -> 'Channel':
        """Creates a voice channel in the guild."""
        from .channel import Channel, ChannelType
        
        category_id = None
        if category:
            if isinstance(category, str):
                category_id = category
            elif hasattr(category, 'id'):
                category_id = category.id

        data = await self._state.http.create_channel(self.id, name, ChannelType.GUILD_VOICE, category_id)
        channel = Channel(self._state, data, self)
        self._channels[channel.id] = channel
        return channel

    async def create_category(self, name: str) -> 'Channel':
        """Creates a category in the guild."""
        from .channel import Channel
        data = await self._state.http.create_category(self.id, name)
        # Category returns data similar to channel
        channel = Channel(self._state, data, self)
        self._channels[channel.id] = channel
        return channel

    # Moderation
    async def kick(self, user: Snowflake, *, reason: Optional[str] = None):
        """Kicks a user from the guild."""
        await self._state.http.kick_member(self.id, user.id)

    async def ban(self, user: Snowflake, *, reason: Optional[str] = None):
        """Bans a user from the guild."""
        await self._state.http.ban_member(self.id, user.id)

    async def unban(self, user: Snowflake, *, reason: Optional[str] = None):
        """Unbans a user from the guild."""
        await self._state.http.unban_member(self.id, user.id)

    async def bans(self):
        """Returns a list of bans."""
        return await self._state.http.get_bans(self.id)

    async def create_role(self, name: str, color: int = 0, permissions=None, hoist=False, mentionable=False):
        """Creates a new role."""
        # Backend creates a default role, so we must edit it immediately
        data = await self._state.http.create_role(self.id, name)
        role = Role(self._state, data, self)
        self._roles[role.id] = role
        
        # Apply parameters
        await role.edit(
            name=name,
            color=color,
            permissions=permissions,
            hoist=hoist,
            mentionable=mentionable
        )
        return role

    def __repr__(self):
        return f'<Guild id={self.id} name={self.name!r} chunked={len(self.members)}/{self.member_count}>'



"""
velmu.models
~~~~~~~~~~~~

Velmu Models Package.
"""
from .base import BaseModel
from .user import User, Member
from .guild import Guild, Role
from .channel import Channel, Message, PartialMessageable
from .interaction import Interaction
from .invite import Invite
from .embed import Embed

__all__ = [
    'BaseModel',
    'User', 'Member',
    'Guild', 'Role',
    'Channel', 'Message', 'PartialMessageable',
    'Interaction',
    'Invite',
    'Embed'
]



"""
velmu.utils
~~~~~~~~~~~

Helper functions and utilities.
"""
import datetime
import warnings

def parse_time(timestamp: str) -> datetime.datetime:
    """Parses an ISO 8601 timestamp."""
    if timestamp:
        return datetime.datetime.fromisoformat(timestamp.replace('Z', '+00:00'))
    return datetime.datetime.utcnow()

def snowflake_time(id: int) -> datetime.datetime:
    """(Placeholder) Returns creation time from a Velmu Snowflake."""
    # Velmu uses UUIDs currently, but if we migrate to Snowflakes:
    # return datetime.datetime.utcfromtimestamp(((id >> 22) + 1420070400000) / 1000)
    return datetime.datetime.utcnow()

def get(iterable, **attrs):
    """A helper that returns the first element in the iterable that meets
    all the traits passed in ``attrs``.

    Args:
        iterable (Iterable): The iterable to search through.
        **attrs (Any): Keyword arguments that denote attributes to search with.
    """
    for item in iterable:
        if all(getattr(item, attr) == value for attr, value in attrs.items()):
            return item
    return None

def find(predicate, iterable):
    """A helper to return the first element in the iterable that satisfies
    the predicate.

    Args:
        predicate (Callable): A callable that returns a boolean.
        iterable (Iterable): The iterable to search through.
    """
    for element in iterable:
        if predicate(element):
            return element
    return None



"""
Velmu SDK

The official Python SDK for building bots on Velmu.

Quick Start
-----------
```python
import velmu
from velmu.ext import commands

intents = velmu.Intents.default()
intents.message_content = True

bot = commands.Bot(command_prefix='+', intents=intents)

@bot.event
async def on_ready():
    print(f'Bot ready: {bot.user.username}')

@bot.command()
async def ping(ctx):
    await ctx.send('Pong!')

bot.run('YOUR_TOKEN')
```
"""

# Core
from .client import Client
from .flags import Intents
from .api import HTTPClient
# Models
from .models import (
    BaseModel,
    User,
    Member,
    Guild,
    Channel,
    Message,
    Interaction,
    PartialMessageable,
    Embed,
    Role
)

# Exceptions
from .errors import (
    VelmuException,
    HTTPException,
    Forbidden,
    NotFound,
    RateLimited,
    VelmuServerError,
    LoginFailure,
    GatewayError,
    CommandError,
    # MissingRequiredArgument, # These are usually in ext.commands, but okay if aliased
)

# Alias old names for compatibility if needed, or stick to new structure
ServerError = VelmuServerError

__title__ = 'velmu'
__author__ = 'Velmu'
__license__ = 'MIT'
__copyright__ = 'Copyright 2024-2026 Velmu'
from ._version import __version__
print(f"📦 Velmu SDK Loaded: v{__version__} from {__file__}")


__all__ = [
    # Core
    'Client',
    'Intents',
    'Embed',
    'HTTPClient',
    
    # Models
    'BaseModel',
    'User',
    'Member',
    'Guild',
    'Channel',
    'Message',
    'Interaction',
    'PartialMessageable',
    'Role',
    
    # Exceptions
    'VelmuException',
    'HTTPException',
    'Forbidden',
    'NotFound',
    'RateLimited',
    'ServerError',
    'LoginFailure',
    'GatewayError',
    'CommandError',
    'MissingRequiredArgument',
    'BadArgument',
    'CheckFailure',
    'CommandNotFound',
    'CommandOnCooldown',
]



from ...models import Member, User, Channel, Role, Guild, Message, Invite


class Converter:
    async def convert(self, ctx, argument):
        raise NotImplementedError('Derived classes must implement convert.')

class MemberConverter(Converter):
    async def convert(self, ctx, argument):
        if not ctx.guild:
            raise Exception("MemberConverter only works in guilds.")

        user_id = None
        # 1. Check ID (UUID support: length check only, no isdigit)
        if len(argument) > 5 and not argument.startswith('<'):
             user_id = argument
        elif argument.startswith('<@') and argument.endswith('>'):
            # <@123> or <@!123>
            content = argument[2:-1]
            if content.startswith('!'):
                content = content[1:]
            if '|' in content: # <@123|username> style
                 content = content.split('|')[0]
            user_id = content
        
        if user_id:
             # Try Cache
             member = ctx.guild.get_member(user_id)
             if member:
                 return member
             
             # Try API
             try:
                 return await ctx.guild.fetch_member(user_id)
             except Exception:
                 # If explicit ID/Mention passed but failed, we might still want to try name matching if ID was not really an ID?
                 # No, if it looks like an ID/Mention, we commit to that.
                 pass

        # 3. Check Username / Nickname (Name Matching)
        arg_lower = argument.lower()
        # Remove leading @ if present (e.g. "@User")
        if arg_lower.startswith('@'):
            arg_lower = arg_lower[1:]
            
        for member in ctx.guild.members:
            if member.username.lower() == arg_lower or (member.nick and member.nick.lower() == arg_lower):
                return member
                
        raise BadArgument(f'Member "{argument}" not found. Please use ID, Mention or Name.')

class ChannelConverter(Converter):
    async def convert(self, ctx, argument):
        if not ctx.guild:
             raise Exception("ChannelConverter only works in guilds.")
        channel_id = None
        # 1. Check ID
        if len(argument) > 5 and argument.isdigit():
            channel_id = argument
        # 2. Check Mention <#123>
        if argument.startswith('<#') and argument.endswith('>'):
            channel_id = argument[2:-1]
        # 2. UUID or ID (Lax check for Velmu UUIDs)
        elif len(argument) > 5:
            channel_id = argument

        if channel_id:
            # Try Cache
            channel = ctx.guild.channels.get(channel_id) if ctx.guild else ctx.bot.get_channel(channel_id)
            if channel: return channel
            
            # Try Fetch
            try:
                # If in guild, fetch from guild to ensure context? 
                # Actually bot.fetch_channel is fine
                return await ctx.bot.fetch_channel(channel_id)
            except:
                # Catch api errors
                pass
                
        # 3. Name Match (in Guild)
        if ctx.guild:
            for channel in ctx.guild.channels.values():
                if channel.name == argument:
                    return channel

        raise BadArgument(f'Channel "{argument}" not found.')

class UserConverter(Converter):
    async def convert(self, ctx, argument):
        user_id = None
        # 1. Mention
        if argument.startswith('<@') and argument.endswith('>'):
            content = argument[2:-1]
            if content.startswith('!'): content = content[1:]
            user_id = content.split('|')[0]
        # 2. UUID or ID
        elif len(argument) > 5:
            user_id = argument

        if user_id:
            # Try Cache
            user = ctx.bot.get_user(user_id)
            if user: return user
            # Try Fetch
            try:
                return await ctx.bot.fetch_user(user_id)
            except:
                pass
        
        raise BadArgument(f'User "{argument}" not found. Use ID or Mention.')

class RoleConverter(Converter):
    async def convert(self, ctx, argument):
        if not ctx.guild:
            raise Exception("RoleConverter only works in guilds.")

        role_id = None
        
        # 1. Mention <@&ID>
        if argument.startswith('<@&') and argument.endswith('>'):
            role_id = argument[3:-1]
        # 2. UUID or ID
        elif len(argument) > 5:
            role_id = argument

        if role_id:
            role = ctx.guild.get_role(role_id)
            if role: return role
            
            # Roles should be in cache if guild is loaded. 
            # If not, try fetching all roles to refresh cache
            try:
                # print(f"[DEBUG] Fetching roles for guild {ctx.guild.id}...")
                roles = await ctx.guild.fetch_roles()
                # print(f"[DEBUG] Fetched {len(roles)} roles. IDs: {[r.id for r in roles]}")
                role = ctx.guild.get_role(role_id)
                if role: return role
            except Exception as e:
                # print(f"[DEBUG] fetch_roles failed: {e}") 
                pass
        
        # 3. Name Match
        for role in ctx.guild.roles.values():
            if role.name == argument:
                return role
                
        raise BadArgument(f'Role "{argument}" not found.')

class GuildConverter(Converter):
    async def convert(self, ctx, argument):
        guild_id = None
        if len(argument) > 5:
            guild_id = argument
            
        if guild_id:
            guild = ctx.bot.get_guild(guild_id)
            if guild: return guild
            try:
                return await ctx.bot.fetch_guild(guild_id)
            except:
                pass
                
        raise BadArgument(f'Guild "{argument}" not found.')

class MessageConverter(Converter):
    async def convert(self, ctx, argument):
        # Format: MessageID or ChannelID-MessageID or Link
        # Only support MessageID (current channel) or ChannelID-MessageID 
        
        message_id = argument
        channel = ctx.channel
        
        # print(f"[DEBUG] Converting Message: {argument} in channel {channel.id}")

        # Try fetch from channel
        try:
            msg = await channel.fetch_message(message_id)
            # print(f"[DEBUG] Message found: {msg.id}")
            return msg
        except Exception as e:
            # print(f"[DEBUG] fetch_message failed: {e}")
            raise BadArgument(f'Message "{argument}" not found.')

class InviteConverter(Converter):
    async def convert(self, ctx, argument):
        # Extract code from URL or raw code
        code = argument
        if 'velmu/invite/' in code:
            code = code.split('invite/')[-1]
            
        try:
            return await ctx.bot.fetch_invite(code)
        except:
            raise BadArgument(f'Invite "{argument}" not found or invalid.')



"""
velmu.models.invite
~~~~~~~~~~~~~~~~~~~

Invite model.
"""
from typing import Optional, TYPE_CHECKING
from .base import BaseModel
from .user import User
from ..utils import parse_time

if TYPE_CHECKING:
    from ..state import ConnectionState
    from .guild import Guild
    from .channel import Channel

class Invite(BaseModel):
    """Represents a Velmu Invite."""
    
    __slots__ = ('_state', 'code', 'guild', 'channel', 'url', 'inviter', 'uses', 'max_uses', 'max_age', 'temporary', 'created_at')
    
    def __init__(self, state: 'ConnectionState', data: dict):
        self._state = state
        self.code = data.get('code')
        self.url = data.get('url')
        self._update(data)
        
    def _update(self, data: dict):
        self.uses = data.get('uses')
        self.max_uses = data.get('maxUses')
        self.max_age = data.get('maxAge')
        self.temporary = data.get('temporary', False)
        self.created_at = parse_time(data.get('createdAt'))
        
        inviter_data = data.get('inviter')
        self.inviter = User(self._state, inviter_data) if inviter_data else None
        
        # Guild/Channel might be partial or full depending on payload
        # For now we don't deeply parse them unless needed
        self.guild = data.get('guild') # TODO: parse properly if needed
        self.channel = data.get('channel') 



from datetime import datetime

class Embed:
    def __init__(self, **kwargs):
        self.title = kwargs.get('title')
        self.description = kwargs.get('description')
        self.url = kwargs.get('url')
        self.color = kwargs.get('color')
        self.timestamp = kwargs.get('timestamp')
        self.html = kwargs.get('html')
        self.fields = []
        
        # Ownership fields
        self.owner_id = kwargs.get('owner_id')  # User ID who can interact
        self.owner_username = kwargs.get('owner_username')  # Display name for overlay
        self.public = kwargs.get('public', True if not kwargs.get('owner_id') else False)
        
        self.footer = None
        self.image = None
        self.thumbnail = None
        self.author = None

    def set_owner(self, user_id, username=None):
        """Restrict interactions to a specific user"""
        self.owner_id = user_id
        self.owner_username = username
        self.public = False
        return self

    def set_public(self, is_public=True):
        """Allow anyone to interact"""
        self.public = is_public
        return self

    def set_footer(self, text, icon_url=None):
        self.footer = {'text': text, 'icon_url': icon_url}
        return self

    def set_image(self, url):
        self.image = {'url': url}
        return self

    def set_thumbnail(self, url):
        self.thumbnail = {'url': url}
        return self

    def set_author(self, name, url=None, icon_url=None):
        self.author = {'name': name, 'url': url, 'icon_url': icon_url}
        return self

    def add_field(self, name, value, inline=True):
        self.fields.append({'name': name, 'value': value, 'inline': inline})
        return self

    def to_dict(self):
        result = {
            'title': self.title,
            'description': self.description,
            'url': self.url,
            'color': self.color,
            'fields': self.fields,
            'public': self.public
        }

        if self.owner_id:
            result['ownerId'] = self.owner_id
            result['ownerUsername'] = self.owner_username

        if self.html:
            result['html'] = self.html
        
        if self.timestamp:
            result['timestamp'] = self.timestamp.isoformat() if isinstance(self.timestamp, datetime) else self.timestamp

        if self.footer: result['footer'] = self.footer
        if self.image: result['image'] = self.image['url'] if isinstance(self.image, dict) else self.image
        if self.thumbnail: result['thumbnail'] = self.thumbnail['url'] if isinstance(self.thumbnail, dict) else self.thumbnail
        if self.author: result['author'] = self.author
        
        return result




"""
velmu.client
~~~~~~~~~~~~

The main Client class.
"""
import asyncio
import logging
import signal
import sys
import aiohttp
from typing import Optional, List, Any, Callable, Coroutine

from .api import HTTPClient
from .gateway import VelmuGateway
from .state import ConnectionState
from .flags import Intents
from .models import User, Guild
from .errors import LoginFailure, GatewayError

_log = logging.getLogger(__name__)

class Client:
    """The client that connects to Velmu.
    
    Attributes
    ----------
    loop : asyncio.AbstractEventLoop
        The event loop.
    intents : Intents
        The intents flags.
    user : Optional[User]
        The connected user.
    """

    def __init__(self, intents: Optional[Intents] = None, **options: Any):
        self.intents = intents or Intents.default()
        
        self.http = HTTPClient()
        self._connection = ConnectionState(dispatch=self.dispatch, http=self.http)
        
        self._gateway: Optional[VelmuGateway] = None
        self._closed = False
        self._ready = asyncio.Event()
        
        # Event handlers storage
        self._events = {}
        self.extra_events = {} # For extensions

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    @property
    def user(self) -> Optional[User]:
        return self._connection.user
    
    @property
    def guilds(self) -> List[Guild]:
        return list(self._connection._guilds.values())

    @property
    def users(self) -> List[User]:
        return list(self._connection._users.values())

    @property
    def latency(self) -> float:
        return self._gateway.latency if self._gateway else 0.0

    # --- Event Registry ---

    def event(self, coro: Callable[..., Coroutine[Any, Any, Any]]) -> Callable[..., Coroutine[Any, Any, Any]]:
        """Decorator to register an event."""
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('Events must be coroutines.')
        
        self._events[coro.__name__] = coro
        _log.debug(f'Registered event: {coro.__name__}')
        return coro

    def dispatch(self, event_name: str, *args: Any, **kwargs: Any):
        """Dispatches an event."""
        _log.debug(f'Dispatching: {event_name}')
        
        # Normalize event name
        if event_name == 'MESSAGE_CREATE':
            normalized_name = 'message'
        elif event_name == 'GUILD_CREATE':
            normalized_name = 'guild_available'
        elif event_name == 'MESSAGE_DELETE':
            normalized_name = 'message_delete'
            # Try to resolve message from cache to pass object instead of dict
            if args and isinstance(args[0], dict):
                payload = args[0]
                message_id = payload.get('id')
                cached_message = self._connection._get_message(message_id) # Access internal
                if cached_message:
                    # Enrich with deleter_id if available in payload
                    cached_message.deleter_id = payload.get('deleterId')
                    args = (cached_message,)
                else:
                    # Pass raw payload if not found
                    pass
        else:
            normalized_name = event_name.lower()

        # Event Parsing
        if normalized_name == 'ready':
            args = ()
        elif normalized_name == 'message':
            # Parse raw dict into Message object
            if args and isinstance(args[0], dict):
                # Use store_message to cache it automatically!
                msg_data = args[0]
                channel_id = msg_data.get('channelId')
                cached_channel = self._connection.get_channel(channel_id)
                if not cached_channel:
                    pass
                else:
                    pass
                    
                message = self._connection.store_message(msg_data)
                args = (message,)
        elif normalized_name == 'guild_available':
             if args and isinstance(args[0], dict):
                # Cache Population
                guild_data = args[0]
                # _log.info(f"Received GUILD_AVAILABLE for {guild_data.get('id')}")
                
                guild = self._connection.store_guild(guild_data)
                
                # Cache Channels (Orphans)
                channels = guild_data.get('channels') or guild_data.get('Channels') or []
                if channels:
                    _log.debug(f"Caching {len(channels)} orphan channels for guild {guild.id}")
                    for channel_data in channels:
                        channel_data['guildId'] = guild.id
                        self._connection.store_channel(channel_data)
                
                # Cache Categories and their Channels
                categories = guild_data.get('categories') or []
                if categories:
                    _log.debug(f"Caching {len(categories)} categories for guild {guild.id}")
                    for cat_data in categories:
                        cat_data['guildId'] = guild.id
                        # Store Category itself as a channel
                        self._connection.store_channel(cat_data)
                        
                        # Store sub-channels
                        sub_channels = cat_data.get('channels') or []
                        # print(f"[SDK DEBUG]   Category {cat_data.get('name')} has {len(sub_channels)} sub-channels")
                        for sub in sub_channels:
                            sub['guildId'] = guild.id
                            sub['categoryId'] = cat_data.get('id') # Ensure linkage
                            self._connection.store_channel(sub)

                # Cache Members (Users)
                members = guild_data.get('members') or [] # members usually contains { userId, user: {...}, ... }
                if members:
                    # _log.info(f"Caching {len(members)} members for guild {guild_data.get('id')}")
                    for member_data in members:
                        # self._connection.store_user(member_data) # OLD WAY
                        self._connection.store_member(guild, member_data) # NEW WAY

                # Cache Presences
                presences = guild_data.get('presences') or []
                if presences:
                    _log.info(f"Caching {len(presences)} presences for guild {guild.id}")
                    for p in presences:
                        # p: { user: {id: ...}, status: ... }
                        user_id = p.get('user', {}).get('id')
                        status = p.get('status')
                        if user_id:
                            member = guild.get_member(user_id)
                            if member:
                                member.status = status
                            
                            # Also update global user
                            user = self._connection.get_user(user_id)
                            if user:
                                user.status = status

                args = (guild,)

        elif normalized_name == 'message_delete':
            # args[0] is likely { id: "...", channelId: "...", guildId: "..." }
            # We pass it as is, or wrap it?
            # Let's wrap it in a lightweight object or just pass dict for now to be safe
            pass

        # --- Modular Event Handling ---
        
        elif normalized_name in ('channel_create', 'channel_update'):
             if args and isinstance(args[0], dict):
                channel = self._connection.parse_channel_create(args[0])
                args = (channel,) if channel else ()
                
        elif normalized_name == 'channel_delete':
             if args and isinstance(args[0], dict):
                channel = self._connection.parse_channel_delete(args[0])
                args = (channel,) if channel else (args[0],)

        elif normalized_name == 'guild_member_add':
             if args and isinstance(args[0], dict):
                member = self._connection.parse_guild_member_add(args[0])
                if member:
                    # FIX: Map to 'member_join' to match Discord.py convention
                    self.dispatch('member_join', member)
                else:
                     import logging
                     logging.getLogger('velmu.client').error(f"Failed to parse Member from GUILD_MEMBER_ADD: {args[0]}")
                
                 # Prevent double dispatch of 'guild_member_add' with empty args
                return 

        elif normalized_name == 'guild_member_remove':
             if args and isinstance(args[0], dict):
                member = self._connection.parse_guild_member_remove(args[0])
                if member:
                    # FIX: Map to 'member_remove'
                    self.dispatch('member_remove', member)
                return 

        elif normalized_name == 'guild_ban_add':
             if args and isinstance(args[0], dict):
                res = self._connection.parse_guild_ban_add(args[0])
                if res:
                    guild, user = res
                    self.dispatch('member_ban', guild, user)
                return

        elif normalized_name == 'guild_ban_remove':
             if args and isinstance(args[0], dict):
                res = self._connection.parse_guild_ban_remove(args[0])
                if res:
                    guild, user = res
                    self.dispatch('member_unban', guild, user)
                return

        elif normalized_name == 'guild_role_create':
             if args and isinstance(args[0], dict):
                role = self._connection.parse_guild_role_create(args[0])
                args = (role,) if role else ()

        elif normalized_name == 'guild_role_update':
             if args and isinstance(args[0], dict):
                role = self._connection.parse_guild_role_update(args[0])
                args = (role,) if role else ()

        elif normalized_name == 'guild_role_delete':
             if args and isinstance(args[0], dict):
                role = self._connection.parse_guild_role_delete(args[0])
                args = (role,) if role else (args[0],)

        elif normalized_name == 'guild_roles_position_update':
             if args and isinstance(args[0], dict):
                roles = self._connection.parse_guild_roles_position_update(args[0])
                # Dispatch 'guild_role_update' for each? Or a bulk event?
                # Discord.py usually dispatches role_update for each.
                # Let's dispatch a custom 'guild_roles_position_update' with list
                args = (roles,) if roles else ()

        elif normalized_name == 'presence_update':
             if args and isinstance(args[0], dict):
                user = self._connection.parse_presence_update(args[0])
                args = (user,) if user else (args[0],)

        elif normalized_name == 'channel_update_permissions':
             if args and isinstance(args[0], dict):
                 channel = self._connection.parse_channel_update_permissions(args[0])
                 # Just trigger generic channel_update or specific?
                 # channel_update is fine
                 args = (channel,) if channel else ()

        elif normalized_name == 'message_reaction_add':
             if args and isinstance(args[0], dict):
                reaction, user = self._connection.parse_message_reaction_add(args[0])
                if reaction:
                    self.dispatch('reaction_add', reaction, user)
                return

        elif normalized_name == 'message_reaction_remove':
             normalized_name = 'reaction_remove'
             pass
        
        # --- Dispatch to Waiters (for wait_for) ---
        if hasattr(self, '_waiters') and normalized_name in self._waiters:
            waiters = self._waiters[normalized_name]
            # Copy list to allow removal during iteration
            for i, (future, check) in enumerate(waiters[:]):
                try:
                    # Check predicate
                    if check(*args):
                        future.set_result(args[0] if len(args) == 1 else args)
                        waiters.remove((future, check))
                except Exception as e:
                    future.set_exception(e)
                    if (future, check) in waiters:
                        waiters.remove((future, check))

        method_name = 'on_' + normalized_name
        
        # Internal listeners (from Cogs/Decorator)
        method = self._events.get(method_name)
        if method:
            self.loop.create_task(method(*args, **kwargs))
        else:
            # Class listeners (subclassing Client) - ONLY if no override
            try:
                class_method = getattr(self, method_name)
                if class_method:
                     self.loop.create_task(class_method(*args, **kwargs))
            except AttributeError:
                pass

    async def wait_for(self, event: str, *, check: Optional[Callable[..., bool]] = None, timeout: Optional[float] = None) -> Any:
        """Waits for a WebSocket event to be dispatched."""
        future = self.loop.create_future()
        if check is None:
            def _check(*args):
                return True
            check = _check

        event_name = event.lower()
        
        if not hasattr(self, '_waiters'):
            self._waiters = {}

        if event_name not in self._waiters:
            self._waiters[event_name] = []
        
        self._waiters[event_name].append((future, check))
        
        try:
            return await asyncio.wait_for(future, timeout=timeout)
        except asyncio.TimeoutError:
            if event_name in self._waiters:
                try:
                    self._waiters[event_name].remove((future, check))
                except ValueError:
                    pass
            raise


    async def on_error(self, event_method: str, *args: Any, **kwargs: Any):
        """Default error handler."""
        print(f"Error in {event_method}:", file=sys.stderr)
        import traceback
        traceback.print_exc()

    # --- Lifecycle ---

    async def login(self, token: str):
        """Logs in the client."""
        _log.info('Logging in...')
        await self.http.static_login(token)
        
        try:
            data = await self.http.get_me()
            self._connection.user = self._connection.store_user(data)
        except Exception as e:
            raise LoginFailure(f"Login failed: {e}")

    async def connect(self, reconnect: bool = True):
        """Connects to the Gateway."""
        if self._closed:
            return
            
        from .config import GATEWAY_URL
        self._gateway = VelmuGateway(self._connection, self.http._token, self.intents)
        
        try:
            await self._gateway.connect(GATEWAY_URL)
        except Exception as e:
            await self.close()
            raise GatewayError(f"Connection failed: {e}")

    async def start(self, token: str, *, reconnect: bool = True):
        """Starts the client."""
        await self.login(token)
        await self.connect(reconnect=reconnect)

    def run(self, token: str, **kwargs: Any):
        """Blocking call to run the bot."""
        async def runner():
            try:
                await self.start(token, **kwargs)
            finally:
                if not self._closed:
                    await self.close()

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Handle Ctrl+C
        try:
            # Only works if loop is running? No, add_signal_handler works on loop object.
            # But get_running_loop() failed in self.loop property.
            # So we use our local 'loop' variable.
            loop.add_signal_handler(signal.SIGINT, lambda: asyncio.create_task(self.close()))
            loop.add_signal_handler(signal.SIGTERM, lambda: asyncio.create_task(self.close()))
        except (NotImplementedError, AttributeError):
            pass # Windows

        try:
            loop.run_until_complete(runner())
        except KeyboardInterrupt:
            pass
        finally:
            _log.info("Client cleanup...")
            # loop.close() # Optional, strict cleanup

    async def close(self):
        """Closes the client."""
        if self._closed:
            return
            
        self._closed = True
        
        if self._gateway:
            await self._gateway.close()
            
        await self.http.close()
        _log.info("Client closed.")

    async def wait_until_ready(self):
        await self._ready.wait()

    async def __aenter__(self):
        # Ensure HTTP session is created if we are entering context
        if self.http._session is None:
             self.http._session = aiohttp.ClientSession()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if not self._closed:
            await self.close()
        
    # --- Methods ---
    
    def get_channel(self, id: str):
        return self._connection.get_channel(id)

    def get_guild(self, id: str):
        return self._connection.get_guild(id)

    def get_user(self, id: str):
        return self._connection.get_user(id)
        
    async def fetch_user(self, id: str):
        data = await self.http.get_user(id)
        return self._connection.store_user(data)
        
    async def fetch_guild(self, id: str):
        data = await self.http.get_guild(id)
        return self._connection.store_guild(data)
        
    async def fetch_channel(self, id: str):
        data = await self.http.get_channel(id)
        return self._connection.store_channel(data)



"""
velmu.gateway
~~~~~~~~~~~~~

WebSocket Gateway connection manager.
"""
import asyncio
import logging
import time
import socketio
import aiohttp
from typing import Optional, TYPE_CHECKING
from .errors import GatewayError

if TYPE_CHECKING:
    from .client import Client
    from .state import ConnectionState
    from .flags import Intents

_log = logging.getLogger(__name__)

class GatewayOp:
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    HELLO = 10
    HEARTBEAT_ACK = 11
    RESUME = 6

class VelmuGateway:
    """Manages the WebSocket connection to the Velmu Gateway."""

    def __init__(self, state: 'ConnectionState', token: str, intents: 'Intents'):
        self.state = state
        self.token = token
        self.intents = intents
        
        # Socket Config
        self._sio = socketio.AsyncClient(
            logger=False, 
            engineio_logger=False,
            reconnection=True,
            reconnection_attempts=0, # Infinite
            reconnection_delay=1,
            reconnection_delay_max=5
        )
        
        # Connection State
        self._heartbeat_task: Optional[asyncio.Task] = None
        self._connected = False
        self._closing = False
        self._session_id: Optional[str] = None
        self._sequence: Optional[int] = None
        
        # Latency State
        self._last_heartbeat_send: float = 0.0
        self._latency: float = float('inf')
        
        # SIO Globals
        self._sio.on('connect', self.on_connect, namespace='/gateway-bot')
        self._sio.on('disconnect', self.on_disconnect, namespace='/gateway-bot')
        self._sio.on('payload', self.on_payload, namespace='/gateway-bot')
        self._sio.on('connect_error', self.on_connect_error, namespace='/gateway-bot')

    @property
    def latency(self) -> float:
        """Latency in seconds."""
        return self._latency

    async def connect(self, url: str):
        """Connect to the Gateway."""
        _log.info(f'Connecting to Gateway: {url}')
        try:
            await self._sio.connect(url, namespaces=['/gateway-bot'], transports=['websocket', 'polling'])
            await self._sio.wait()
        except asyncio.CancelledError:
            await self.close()
        except Exception as e:
            _log.error(f"Gateway connection failed: {e}")
            raise GatewayError(message=str(e))

    async def close(self):
        """Close the connection."""
        self._closing = True
        if self._heartbeat_task:
            self._heartbeat_task.cancel()
        
        if self._sio.connected:
            await self._sio.disconnect()
        
        _log.info("Gateway closed.")

    async def send(self, op: int, data: dict):
        """Send a payload to the Gateway."""
        payload = {'op': op, 'd': data}
        await self._sio.emit('payload', payload, namespace='/gateway-bot')

    # --- Events ---

    async def on_connect(self):
        self._connected = True
        _log.info('Gateway connection established.')

    async def on_disconnect(self):
        self._connected = False
        if not self._closing:
            _log.warning('Gateway disconnected unexpectedly. Reconnecting...')
            # SocketIO handles reconnection, but we might want to clear tasks?
            if self._heartbeat_task:
                self._heartbeat_task.cancel()
        else:
            _log.info('Gateway disconnected.')

    async def on_connect_error(self, data):
        _log.error(f'Gateway Connection Error: {data}')

    async def on_payload(self, data):
        op = data.get('op')
        d = data.get('d')
        t = data.get('t')
        
        # Update sequence
        if data.get('s'):
            self._sequence = data.get('s')

        if op == GatewayOp.HELLO:
            interval = d['heartbeat_interval'] / 1000
            self._heartbeat_task = asyncio.create_task(self.heartbeat_loop(interval))
            
            # Resume or Identify
            if self._session_id and self._sequence:
                _log.info('Attempting to RESUME session...')
                await self.resume()
            else:
                _log.info('Identifying new session...')
                await self.identify()
        
        elif op == GatewayOp.HEARTBEAT_ACK:
            # Calculate Latency
            ack_time = time.perf_counter()
            self._latency = ack_time - self._last_heartbeat_send
            # _log.debug(f'Heartbeat ACK. Latency: {self._latency*1000:.2f}ms')
            
        elif op == GatewayOp.DISPATCH:
            if t == 'READY':
                self._session_id = d.get('sessionId')
                _log.info(f'Session Ready: {self._session_id}')
                # Explicitly set status online after Ready
                await self.update_presence(status='online')
            self.state.dispatch(t, d)
            
        elif op == 7: # RECONNECT request
            _log.info('Gateway requested RECONNECT.')
            await self.close() # Will trigger client loop to restart? No, we need logic.
            # Actually, socketio auto-reconnect handles it if we just disconnect logic but keep socket retry?
            # Or we strictly follow Discord op 7 -> Resume.
            await self.resume()
            
        elif op == 9: # INVALID SESSION
            _log.warning('Session Invalid. Re-identifying...')
            self._session_id = None
            self._sequence = None
            await self.identify()

    async def identify(self):
        """Send IDENTIFY payload."""
        payload = {
            'token': self.token,
            'intents': self.intents.value,
            'properties': {
                'os': 'linux',
                'browser': 'velmu.py',
                'device': 'velmu.py'
            },
            'presence': {
                'status': 'online',
                'afk': False
            }
        }
        await self.send(GatewayOp.IDENTIFY, payload)
        
    async def resume(self):
        """Send RESUME payload."""
        payload = {
            'token': self.token,
            'sessionId': self._session_id,
            'seq': self._sequence
        }
        await self.send(GatewayOp.RESUME, payload)

    async def heartbeat_loop(self, interval: float):
        _log.debug(f'Heartbeat started (interval: {interval}s)')
        try:
            while True:
                self._last_heartbeat_send = time.perf_counter()
                await self.send(GatewayOp.HEARTBEAT, {'s': self._sequence})
                await asyncio.sleep(interval)
        except asyncio.CancelledError:
            pass

    async def update_presence(self, status: str = 'online', afk: bool = False):
        """Updates the presence status."""
        payload = {
            'since': None,
            'activities': [],
            'status': status,
            'afk': afk
        }
        await self.send(3, payload) # Opcode 3: Prevence Update



from .core import Context

def check(predicate):
    def decorator(func):
        if not hasattr(func, 'checks'):
            func.checks = []
        func.checks.append(predicate)
        return func
    return decorator

def has_permissions(**perms):
    def predicate(ctx: Context):
        # TODO: Check actual permissions in ctx.author.permissions_in(ctx.channel)
        # For now, return True or check simple flags if available in User object
        return True 
    return check(predicate)

def is_owner():
    def predicate(ctx: Context):
        # return ctx.author.id == ctx.bot.owner_id
        return True
    return check(predicate)



"""
Velmu Intent Flags

Bitfield flags for Gateway intents.
"""


class BaseFlags:
    """Base class for bitfield flags."""
    
    def __init__(self, value=0):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.value)

    def __iter__(self):
        for name, value in self.__class__.__dict__.items():
            if isinstance(value, int) and not name.startswith('_'):
                yield name, (self.value & value) == value
    
    def __repr__(self):
        enabled = [name for name, enabled in self if enabled]
        return f'<{self.__class__.__name__} value={self.value} enabled={enabled}>'

    def _set_flag(self, flag, value):
        if value:
            self.value |= flag
        else:
            self.value &= ~flag


class Intents(BaseFlags):
    """Gateway intents for subscribing to events.
    
    Intents control which events your bot receives from the Gateway.
    Some intents are **privileged** and must be enabled in the Developer Portal.
    
    Privileged Intents
    ------------------
    - GUILD_MEMBERS (1 << 1): Required for member join/leave events
    - GUILD_PRESENCES (1 << 8): Required for presence updates
    - MESSAGE_CONTENT (1 << 15): Required to read message content
    
    Example
    -------
    ```python
    # Default intents (most common use case)
    intents = Intents.default()
    
    # Enable a privileged intent
    intents.message_content = True
    intents.members = True
    
    # All intents (requires all privileged intents enabled in portal)
    intents = Intents.all()
    ```
    """
    
    # Guild-related
    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1  # PRIVILEGED
    GUILD_MODERATION = 1 << 2
    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8  # PRIVILEGED
    
    # Message-related
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    
    # DM-related
    DIRECT_MESSAGES = 1 << 12
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    DIRECT_MESSAGE_TYPING = 1 << 14
    
    # Content
    MESSAGE_CONTENT = 1 << 15  # PRIVILEGED
    
    # Scheduled Events
    GUILD_SCHEDULED_EVENTS = 1 << 16
    
    # Auto Moderation
    AUTO_MODERATION_CONFIGURATION = 1 << 20
    AUTO_MODERATION_EXECUTION = 1 << 21

    # =========================================
    # Factory Methods
    # =========================================

    @classmethod
    def default(cls) -> 'Intents':
        """Create intents with common non-privileged intents enabled.
        
        Includes: GUILDS, GUILD_MESSAGES, GUILD_MESSAGE_REACTIONS,
        DIRECT_MESSAGES, DIRECT_MESSAGE_REACTIONS
        """
        self = cls(0)
        self.value = (
            cls.GUILDS |
            cls.GUILD_MESSAGES |
            cls.GUILD_MESSAGE_REACTIONS |
            cls.DIRECT_MESSAGES |
            cls.DIRECT_MESSAGE_REACTIONS
        )
        return self

    @classmethod
    def all(cls) -> 'Intents':
        """Create intents with ALL intents enabled.
        
        ⚠️ Warning: This includes privileged intents that must be
        enabled in the Developer Portal.
        """
        self = cls(0)
        all_bits = 0
        for name, value in cls.__dict__.items():
            if isinstance(value, int) and not name.startswith('_'):
                all_bits |= value
        self.value = all_bits
        return self
    
    @classmethod
    def none(cls) -> 'Intents':
        """Create intents with no intents enabled."""
        return cls(0)
    
    @classmethod
    def privileged(cls) -> 'Intents':
        """Create intents with only privileged intents enabled."""
        self = cls(0)
        self.value = cls.GUILD_MEMBERS | cls.GUILD_PRESENCES | cls.MESSAGE_CONTENT
        return self

    # =========================================
    # Guild Properties
    # =========================================

    @property
    def guilds(self) -> bool:
        """Whether GUILDS intent is enabled."""
        return (self.value & self.GUILDS) == self.GUILDS
    
    @guilds.setter
    def guilds(self, v: bool):
        self._set_flag(self.GUILDS, v)

    @property
    def members(self) -> bool:
        """Whether GUILD_MEMBERS intent is enabled. (PRIVILEGED)"""
        return (self.value & self.GUILD_MEMBERS) == self.GUILD_MEMBERS
    
    @members.setter
    def members(self, v: bool):
        self._set_flag(self.GUILD_MEMBERS, v)

    @property
    def moderation(self) -> bool:
        """Whether GUILD_MODERATION intent is enabled."""
        return (self.value & self.GUILD_MODERATION) == self.GUILD_MODERATION
    
    @moderation.setter
    def moderation(self, v: bool):
        self._set_flag(self.GUILD_MODERATION, v)

    @property
    def emojis(self) -> bool:
        """Whether GUILD_EMOJIS_AND_STICKERS intent is enabled."""
        return (self.value & self.GUILD_EMOJIS_AND_STICKERS) == self.GUILD_EMOJIS_AND_STICKERS
    
    @emojis.setter
    def emojis(self, v: bool):
        self._set_flag(self.GUILD_EMOJIS_AND_STICKERS, v)

    @property
    def integrations(self) -> bool:
        """Whether GUILD_INTEGRATIONS intent is enabled."""
        return (self.value & self.GUILD_INTEGRATIONS) == self.GUILD_INTEGRATIONS
    
    @integrations.setter
    def integrations(self, v: bool):
        self._set_flag(self.GUILD_INTEGRATIONS, v)

    @property
    def webhooks(self) -> bool:
        """Whether GUILD_WEBHOOKS intent is enabled."""
        return (self.value & self.GUILD_WEBHOOKS) == self.GUILD_WEBHOOKS
    
    @webhooks.setter
    def webhooks(self, v: bool):
        self._set_flag(self.GUILD_WEBHOOKS, v)

    @property
    def invites(self) -> bool:
        """Whether GUILD_INVITES intent is enabled."""
        return (self.value & self.GUILD_INVITES) == self.GUILD_INVITES
    
    @invites.setter
    def invites(self, v: bool):
        self._set_flag(self.GUILD_INVITES, v)

    @property
    def voice_states(self) -> bool:
        """Whether GUILD_VOICE_STATES intent is enabled."""
        return (self.value & self.GUILD_VOICE_STATES) == self.GUILD_VOICE_STATES
    
    @voice_states.setter
    def voice_states(self, v: bool):
        self._set_flag(self.GUILD_VOICE_STATES, v)

    @property
    def presences(self) -> bool:
        """Whether GUILD_PRESENCES intent is enabled. (PRIVILEGED)"""
        return (self.value & self.GUILD_PRESENCES) == self.GUILD_PRESENCES
    
    @presences.setter
    def presences(self, v: bool):
        self._set_flag(self.GUILD_PRESENCES, v)

    # =========================================
    # Message Properties
    # =========================================

    @property
    def guild_messages(self) -> bool:
        """Whether GUILD_MESSAGES intent is enabled."""
        return (self.value & self.GUILD_MESSAGES) == self.GUILD_MESSAGES
    
    @guild_messages.setter
    def guild_messages(self, v: bool):
        self._set_flag(self.GUILD_MESSAGES, v)

    @property
    def guild_reactions(self) -> bool:
        """Whether GUILD_MESSAGE_REACTIONS intent is enabled."""
        return (self.value & self.GUILD_MESSAGE_REACTIONS) == self.GUILD_MESSAGE_REACTIONS
    
    @guild_reactions.setter
    def guild_reactions(self, v: bool):
        self._set_flag(self.GUILD_MESSAGE_REACTIONS, v)

    @property
    def guild_typing(self) -> bool:
        """Whether GUILD_MESSAGE_TYPING intent is enabled."""
        return (self.value & self.GUILD_MESSAGE_TYPING) == self.GUILD_MESSAGE_TYPING
    
    @guild_typing.setter
    def guild_typing(self, v: bool):
        self._set_flag(self.GUILD_MESSAGE_TYPING, v)

    @property
    def dm_messages(self) -> bool:
        """Whether DIRECT_MESSAGES intent is enabled."""
        return (self.value & self.DIRECT_MESSAGES) == self.DIRECT_MESSAGES
    
    @dm_messages.setter
    def dm_messages(self, v: bool):
        self._set_flag(self.DIRECT_MESSAGES, v)

    @property
    def dm_reactions(self) -> bool:
        """Whether DIRECT_MESSAGE_REACTIONS intent is enabled."""
        return (self.value & self.DIRECT_MESSAGE_REACTIONS) == self.DIRECT_MESSAGE_REACTIONS
    
    @dm_reactions.setter
    def dm_reactions(self, v: bool):
        self._set_flag(self.DIRECT_MESSAGE_REACTIONS, v)

    @property
    def dm_typing(self) -> bool:
        """Whether DIRECT_MESSAGE_TYPING intent is enabled."""
        return (self.value & self.DIRECT_MESSAGE_TYPING) == self.DIRECT_MESSAGE_TYPING
    
    @dm_typing.setter
    def dm_typing(self, v: bool):
        self._set_flag(self.DIRECT_MESSAGE_TYPING, v)

    @property
    def message_content(self) -> bool:
        """Whether MESSAGE_CONTENT intent is enabled. (PRIVILEGED)"""
        return (self.value & self.MESSAGE_CONTENT) == self.MESSAGE_CONTENT
    
    @message_content.setter
    def message_content(self, v: bool):
        self._set_flag(self.MESSAGE_CONTENT, v)

    # =========================================
    # Other Properties
    # =========================================

    @property
    def scheduled_events(self) -> bool:
        """Whether GUILD_SCHEDULED_EVENTS intent is enabled."""
        return (self.value & self.GUILD_SCHEDULED_EVENTS) == self.GUILD_SCHEDULED_EVENTS
    
    @scheduled_events.setter
    def scheduled_events(self, v: bool):
        self._set_flag(self.GUILD_SCHEDULED_EVENTS, v)

    @property
    def auto_moderation_configuration(self) -> bool:
        """Whether AUTO_MODERATION_CONFIGURATION intent is enabled."""
        return (self.value & self.AUTO_MODERATION_CONFIGURATION) == self.AUTO_MODERATION_CONFIGURATION
    
    @auto_moderation_configuration.setter
    def auto_moderation_configuration(self, v: bool):
        self._set_flag(self.AUTO_MODERATION_CONFIGURATION, v)

    @property
    def auto_moderation_execution(self) -> bool:
        """Whether AUTO_MODERATION_EXECUTION intent is enabled."""
        return (self.value & self.AUTO_MODERATION_EXECUTION) == self.AUTO_MODERATION_EXECUTION
    
    @auto_moderation_execution.setter
    def auto_moderation_execution(self, v: bool):
        self._set_flag(self.AUTO_MODERATION_EXECUTION, v)



"""
velmu.abc
~~~~~~~~~

Abstract Base Classes for the Velmu SDK.
"""
import abc
import datetime
from .errors import ClientException

class Snowflake(abc.ABC):
    """An abstract base class for objects that have a unique ID."""
    
    __slots__ = ()

    @property
    @abc.abstractmethod
    def id(self) -> str:
        """str: The unique ID of the object."""
        raise NotImplementedError

    @property
    def created_at(self) -> datetime.datetime:
        """datetime.datetime: The time the object was created (if derivable or stored).
        For standard UUIDs this might not be derivable, but we define the interface.
        """
        # Default implementation if not overridden
        return datetime.datetime.utcnow() 

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.id == self.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)


class Messageable(abc.ABC):
    """An abstract base class for models that can send messages.
    
    e.g. Channel, User (DM), Context
    """
    
    __slots__ = ()

    @abc.abstractmethod
    async def _get_channel(self):
        raise NotImplementedError

    async def send(self, content=None, *, embed=None, **kwargs):
        """Sends a message to the destination.
        
        Parameters
        ----------
        content : str
            The content of the message.
        embed : Embed
            The embed to send.
            
        Returns
        -------
        Message
            The sent message.
        """
        channel = await self._get_channel()
        # This assumes channel has a send method, or we use the state/http to send
        # Designing this depends on how we structure models. 
        # Ideally, Messageable models have access to 'state' (http).
        pass

class Connectable(abc.ABC):
    """An interface for objects that need a connection state."""
    __slots__ = ()
    
    @property
    @abc.abstractmethod
    def _state(self):
        raise NotImplementedError
